# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class AuthenticationProvider(object):
    """Super class for all providing known types of authentication mechanism like
    x509 and SAS based authentication.

    :ivar str hostname: Hostname
    :ivar str device_id: Device ID
    :ivar str module_id: Module ID
    """

    def __init__(self, hostname, device_id, module_id=None):
        """Initializer for AuthenticationProvider

        :param str hostname: Hostname
        :param str device_id: Device ID
        :param str module_id: Module ID (optional)
        """
        self.hostname = hostname
        self.device_id = device_id
        self.module_id = module_id


# TODO: Potentially some additional abstract class that defines an abstract .get_current_sas_token()
# in order to enforce sas token retrieval in various sas-affiliated auths (sas, sk)?



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains tools for adapting sync code for use in async coroutines."""

import functools
import logging
import traceback
import azure.iot.device.common.asyncio_compat as asyncio_compat

logger = logging.getLogger(__name__)


def emulate_async(fn):
    """Returns a coroutine function that calls a given function with emulated asynchronous
    behavior via use of mulithreading.

    Can be applied as a decorator.

    :param fn: The sync function to be run in async.
    :returns: A coroutine function that will call the given sync function.
    """

    @functools.wraps(fn)
    async def async_fn_wrapper(*args, **kwargs):
        loop = asyncio_compat.get_running_loop()

        # Run fn in default ThreadPoolExecutor (CPU * 5 threads)
        return await loop.run_in_executor(None, functools.partial(fn, *args, **kwargs))

    return async_fn_wrapper


class AwaitableCallback(object):
    """A sync callback whose completion can be waited upon.
    """

    def __init__(self, return_arg_name=None):
        """Creates an instance of an AwaitableCallback
        """

        # LBYL because this mistake doesn't cause an exception until the callback
        # which is much later and very difficult to trace back to here.
        if return_arg_name and not isinstance(return_arg_name, str):
            raise TypeError("internal error: return_arg_name must be a string")

        loop = asyncio_compat.get_running_loop()
        self.future = asyncio_compat.create_future(loop)

        def wrapping_callback(*args, **kwargs):
            # Use event loop from outer scope, since the threads it will be used in will not have
            # an event loop. future.set_result() and future.set_exception have to be called in an
            # event loop or they do not work.
            if "error" in kwargs and kwargs["error"]:
                exception = kwargs["error"]
            elif return_arg_name:
                if return_arg_name in kwargs:
                    exception = None
                    result = kwargs[return_arg_name]
                else:
                    raise TypeError(
                        "internal error: excepected argument with name '{}', did not get".format(
                            return_arg_name
                        )
                    )
            else:
                exception = None
                result = None

            if exception:
                # Do not use exc_info parameter on logger.error.  This casuses pytest to save the traceback which saves stack frames which shows up as a leak
                logger.error("Callback completed with error {}".format(exception))
                logger.error(traceback.format_exception_only(type(exception), exception))
                loop.call_soon_threadsafe(self.future.set_exception, exception)
            else:
                logger.debug("Callback completed with result {}".format(result))
                loop.call_soon_threadsafe(self.future.set_result, result)

        self.callback = wrapping_callback

    def __call__(self, *args, **kwargs):
        """Calls the callback. Returns the result.
        """
        return self.callback(*args, **kwargs)

    async def completion(self):
        """Awaitable coroutine method that will return once the AwaitableCallback
        has been completed.

        :returns: Result of the callback when it was called.
        """
        return await self.future



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import uuid
import threading
import json
import ssl
from . import transport_exceptions as exceptions
from .pipeline import pipeline_thread
from six.moves import http_client

logger = logging.getLogger(__name__)


class HTTPTransport(object):
    """
    A wrapper class that provides an implementation-agnostic HTTP interface.
    """

    def __init__(self, hostname, server_verification_cert=None, x509_cert=None, cipher=None):
        """
        Constructor to instantiate an HTTP protocol wrapper.

        :param str hostname: Hostname or IP address of the remote host.
        :param str server_verification_cert: Certificate which can be used to validate a server-side TLS connection (optional).
        :param str cipher: Cipher string in OpenSSL cipher list format (optional)
        :param x509_cert: Certificate which can be used to authenticate connection to a server in lieu of a password (optional).
        """
        self._hostname = hostname
        self._server_verification_cert = server_verification_cert
        self._x509_cert = x509_cert
        self._cipher = cipher
        self._ssl_context = self._create_ssl_context()

    def _create_ssl_context(self):
        """
        This method creates the SSLContext object used to authenticate the connection. The generated context is used by the http_client and is necessary when authenticating using a self-signed X509 cert or trusted X509 cert
        """
        logger.debug("creating a SSL context")
        ssl_context = ssl.SSLContext(protocol=ssl.PROTOCOL_TLSv1_2)

        if self._server_verification_cert:
            ssl_context.load_verify_locations(cadata=self._server_verification_cert)
        else:
            ssl_context.load_default_certs()

        if self._cipher:
            try:
                ssl_context.set_ciphers(self._cipher)
            except ssl.SSLError as e:
                # TODO: custom error with more detail?
                raise e

        if self._x509_cert is not None:
            logger.debug("configuring SSL context with client-side certificate and key")
            ssl_context.load_cert_chain(
                self._x509_cert.certificate_file,
                self._x509_cert.key_file,
                self._x509_cert.pass_phrase,
            )

        ssl_context.verify_mode = ssl.CERT_REQUIRED
        ssl_context.check_hostname = True

        return ssl_context

    @pipeline_thread.invoke_on_http_thread_nowait
    def request(self, method, path, callback, body="", headers={}, query_params=""):
        """
        This method creates a connection to a remote host, sends a request to that host, and then waits for and reads the response from that request.

        :param str method: The request method (e.g. "POST")
        :param str path: The path for the URL
        :param Function callback: The function that gets called when this operation is complete or has failed. The callback function must accept an error and a response dictionary, where the response dictionary contains a status code, a reason, and a response string.
        :param str body: The body of the HTTP request to be sent following the headers.
        :param dict headers: A dictionary that provides extra HTTP headers to be sent with the request.
        :param str query_params: The optional query parameters to be appended at the end of the URL.
        """
        # Sends a complete request to the server
        logger.info("sending https request.")
        try:
            logger.debug("creating an https connection")
            connection = http_client.HTTPSConnection(self._hostname, context=self._ssl_context)
            logger.debug("connecting to host tcp socket")
            connection.connect()
            logger.debug("connection succeeded")
            url = "https://{hostname}/{path}{query_params}".format(
                hostname=self._hostname,
                path=path,
                query_params="?" + query_params if query_params else "",
            )
            logger.debug("Sending Request to HTTP URL: {}".format(url))
            logger.debug("HTTP Headers: {}".format(headers))
            logger.debug("HTTP Body: {}".format(body))
            connection.request(method, url, body=body, headers=headers)
            response = connection.getresponse()
            status_code = response.status
            reason = response.reason
            response_string = response.read()

            logger.debug("response received")
            logger.debug("closing connection to https host")
            connection.close()
            logger.debug("connection closed")
            logger.info("https request sent, and response received.")
            response_obj = {"status_code": status_code, "reason": reason, "resp": response_string}
            callback(response=response_obj)
        except Exception as e:
            logger.error("Error in HTTP Transport: {}".format(e))
            callback(
                error=exceptions.ProtocolClientError(
                    message="Unexpected HTTPS failure during connect", cause=e
                )
            )



"""Azure IoT Provisioning Service SDK - Asynchronous

This SDK provides asynchronous functionality for communicating with the Azure Provisioning Service
as a Device.
"""

from .async_provisioning_device_client import ProvisioningDeviceClient

__all__ = ["ProvisioningDeviceClient"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.iot.device.common.pipeline.pipeline_ops_base import PipelineOperation


class SetSymmetricKeySecurityClientOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to use a symmetric key security client.
    Some pipeline stage is expected to extract arguments out of the security client and pass them
    on so an even lower stage can use those arguments to connect.

    This operation is in the group of provisioning operations because security clients are currently
    very provisioning-specific
    """

    def __init__(self, security_client, callback):
        """
        Initializer for SetSecurityClient.

        :param object security_client: The security client object to use to retrieve connection parameters
          which can be used to connect to the service.
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super(SetSymmetricKeySecurityClientOperation, self).__init__(callback=callback)
        self.security_client = security_client


class SetX509SecurityClientOperation(PipelineOperation):
    """
    A PipelineOperation object which contains connection arguments which were retrieved from a
    X509 security client likely by a pipeline stage which handles the
    SetX509SecurityClientOperation operation.

    This operation is in the group of Provisioning operations because the arguments which it accepts are
    very specific to DPS connections and would not apply to other types of client connections
    (such as a Provisioning client).
    """

    def __init__(self, security_client, callback):
        """
        Initializer for SetSecurityClient.

        :param object security_client: The security client object to use to retrieve connection parameters
        which can be used to connect to the service.
        :param Function callback: The function that gets called when this operation is complete or has failed.
        The callback function must accept A PipelineOperation object which indicates the specific operation which
        has completed or failed.
        """
        super(SetX509SecurityClientOperation, self).__init__(callback=callback)
        self.security_client = security_client


class SetProvisioningClientConnectionArgsOperation(PipelineOperation):
    """
    A PipelineOperation object which contains connection arguments which were retrieved from a
    symmetric key or a X509 security client likely by a pipeline stage which handles the
    SetSymmetricKeySecurityClientOperation or SetX509SecurityClientOperation operation.

    This operation is in the group of Provisioning operations because the arguments which it accepts are
    very specific to DPS connections and would not apply to other types of client connections
    (such as a Provisioning client).
    """

    def __init__(
        self,
        provisioning_host,
        registration_id,
        id_scope,
        callback,
        client_cert=None,
        sas_token=None,
    ):
        """
        Initializer for SetProvisioningClientConnectionArgsOperation.
        :param registration_id: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        The registration ID is alphanumeric, lowercase string and may contain hyphens.
        :param id_scope: The ID scope is used to uniquely identify the specific provisioning service the device will
        register through. The ID scope is assigned to a Device Provisioning Service when it is created by the user and
        is generated by the service and is immutable, guaranteeing uniqueness.
        """
        super(SetProvisioningClientConnectionArgsOperation, self).__init__(callback=callback)
        self.provisioning_host = provisioning_host
        self.registration_id = registration_id
        self.id_scope = id_scope
        self.client_cert = client_cert
        self.sas_token = sas_token


class RegisterOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to send a registration request
    to an Device Provisioning Service.

    This operation is in the group of DPS operations because it is very specific to the DPS client.
    """

    def __init__(self, request_payload, registration_id, callback, registration_result=None):
        """
        Initializer for RegisterOperation objects.

        :param request_payload: The request that we are sending to the service
        :param registration_id: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super(RegisterOperation, self).__init__(callback=callback)
        self.request_payload = request_payload
        self.registration_id = registration_id
        self.registration_result = registration_result
        self.retry_after_timer = None
        self.polling_timer = None
        self.provisioning_timeout_timer = None


class PollStatusOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to send a registration request
    to an Device Provisioning Service.

    This operation is in the group of DPS operations because it is very specific to the DPS client.
    """

    def __init__(self, operation_id, request_payload, callback, registration_result=None):
        """
        Initializer for PollStatusOperation objects.

        :param operation_id: The id of the existing operation for which the polling was started.
        :param request_payload: The request that we are sending to the service
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super(PollStatusOperation, self).__init__(callback=callback)
        self.operation_id = operation_id
        self.request_payload = request_payload
        self.registration_result = registration_result
        self.retry_after_timer = None
        self.polling_timer = None
        self.provisioning_timeout_timer = None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from azure.iot.device.common.pipeline.config import BasePipelineConfig

logger = logging.getLogger(__name__)


class IoTHubPipelineConfig(BasePipelineConfig):
    """A class for storing all configurations/options for IoTHub clients in the Azure IoT Python Device Client Library.
    """

    def __init__(self, product_info="", **kwargs):
        """Initializer for IoTHubPipelineConfig which passes all unrecognized keyword-args down to BasePipelineConfig
        to be evaluated. This stacked options setting is to allow for unique configuration options to exist between the
        IoTHub Client and the Provisioning Client, while maintaining a base configuration class with shared config options.

        :param str product_info: A custom identification string for the type of device connecting to Azure IoT Hub.
        """
        super(IoTHubPipelineConfig, self).__init__(**kwargs)
        self.product_info = product_info

        # Now, the parameters below are not exposed to the user via kwargs. They need to be set by manipulating the IoTHubPipelineConfig object.
        # They are not in the BasePipelineConfig because these do not apply to the provisioning client.
        self.blob_upload = False
        self.method_invoke = False



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module defines errors that may be raised from a transport"""

from .chainable_exception import ChainableException


class ConnectionFailedError(ChainableException):
    """
    Connection failed to be established
    """

    pass


class ConnectionDroppedError(ChainableException):
    """
    Previously established connection was dropped
    """

    pass


class UnauthorizedError(ChainableException):
    """
    Authorization was rejected
    """

    pass


class ProtocolClientError(ChainableException):
    """
    Error returned from protocol client library
    """

    pass


class TlsExchangeAuthError(ChainableException):
    """
    Error returned when transport layer exchanges
    result in a SSLCertVerification error.
    """

    pass


class ProtocolProxyError(ChainableException):
    """
    All proxy-related errors.
    TODO : Not sure what to name it here. There is a class called Proxy Error already in Pysocks
    """

    pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from .authentication_provider import AuthenticationProvider

logger = logging.getLogger(__name__)


class X509AuthenticationProvider(AuthenticationProvider):
    """
    An X509 Authentication Provider. This provider uses the certificate and key provided to
    authenticate a device with an Azure IoT Hub instance. X509 Authentication is only supported
    for device identities connecting directly to an Azure IoT hub.
    """

    def __init__(self, x509, hostname, device_id, module_id=None):
        """
        Constructor for X509 Authentication Provider
        :param x509: The X509 object containing certificate, key and passphrase
        :param hostname: The hostname of the Azure IoT hub.
        :param device_id: The device unique identifier as it exists in the Azure IoT Hub device registry.
        :param module_id: The module unique identifier of the device. It is not applicable when dealing with only devices.
        """
        logger.info(
            "Using X509 authentication for {hostname},{device_id},{module_id}".format(
                hostname=hostname, device_id=device_id, module_id=module_id
            )
        )
        super(X509AuthenticationProvider, self).__init__(
            hostname=hostname, device_id=device_id, module_id=module_id
        )
        self._x509 = x509

    def get_x509_certificate(self):
        """
        :return: The x509 certificate, To use the certificate the enrollment object needs to contain
         cert (either the root certificate or one of the intermediate CA certificates).
        """
        return self._x509



"""Azure Provisioning Device Communication Pipeline

This package provides pipeline for use with the Azure Provisioning Device SDK.

INTERNAL USAGE ONLY
"""
from .provisioning_pipeline import ProvisioningPipeline
from .config import ProvisioningPipelineConfig



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module defines exceptions that may be raised from a pipeline"""

from azure.iot.device.common.chainable_exception import ChainableException


class PipelineException(ChainableException):
    """Generic pipeline exception"""

    pass


class OperationCancelled(PipelineException):
    """Operation was cancelled"""

    pass


class OperationError(PipelineException):
    """Error while executing an Operation"""

    pass


class PipelineTimeoutError(PipelineException):
    """
    Pipeline operation timed out
    """

    pass


class PipelineError(PipelineException):
    """Error caused by incorrect pipeline configuration"""

    pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
This module contains user-facing synchronous Provisioning Device Client for Azure Provisioning
Device SDK. This client uses Symmetric Key and X509 authentication to register devices with an
IoT Hub via the Device Provisioning Service.
"""
import logging
from azure.iot.device.common.evented_callback import EventedCallback
from .abstract_provisioning_device_client import AbstractProvisioningDeviceClient
from .abstract_provisioning_device_client import log_on_register_complete
from azure.iot.device.provisioning.pipeline import constant as dps_constant
from .pipeline import exceptions as pipeline_exceptions
from azure.iot.device import exceptions


logger = logging.getLogger(__name__)


def handle_result(callback):
    try:
        return callback.wait_for_completion()
    except pipeline_exceptions.ConnectionDroppedError as e:
        raise exceptions.ConnectionDroppedError(message="Lost connection to IoTHub", cause=e)
    except pipeline_exceptions.ConnectionFailedError as e:
        raise exceptions.ConnectionFailedError(message="Could not connect to IoTHub", cause=e)
    except pipeline_exceptions.UnauthorizedError as e:
        raise exceptions.CredentialError(message="Credentials invalid, could not connect", cause=e)
    except pipeline_exceptions.ProtocolClientError as e:
        raise exceptions.ClientError(message="Error in the IoTHub client", cause=e)
    except Exception as e:
        raise exceptions.ClientError(message="Unexpected failure", cause=e)


class ProvisioningDeviceClient(AbstractProvisioningDeviceClient):
    """
    Client which can be used to run the registration of a device with provisioning service
    using Symmetric Key orr X509 authentication.
    """

    def register(self):
        """
        Register the device with the with the provisioning service

        This is a synchronous call, meaning that this function will not return until the
        registration process has completed successfully or the attempt has resulted in a failure.
        Before returning, the client will also disconnect from the provisioning service.
        If a registration attempt is made while a previous registration is in progress it may
        throw an error.

        :returns: RegistrationResult indicating the result of the registration.
        :rtype: :class:`azure.iot.device.RegistrationResult`

        :raises: :class:`azure.iot.device.exceptions.CredentialError` if credentials are invalid
            and a connection cannot be established.
        :raises: :class:`azure.iot.device.exceptions.ConnectionFailedError` if a establishing a
            connection results in failure.
        :raises: :class:`azure.iot.device.exceptions.ConnectionDroppedError` if connection is lost
            during execution.
        :raises: :class:`azure.iot.device.exceptions.ClientError` if there is an unexpected failure
            during execution.
        """
        logger.info("Registering with Provisioning Service...")

        if not self._provisioning_pipeline.responses_enabled[dps_constant.REGISTER]:
            self._enable_responses()

        register_complete = EventedCallback(return_arg_name="result")

        self._provisioning_pipeline.register(
            payload=self._provisioning_payload, callback=register_complete
        )

        result = handle_result(register_complete)

        log_on_register_complete(result)
        return result

    def _enable_responses(self):
        """Enable to receive responses from Device Provisioning Service.

        This is a synchronous call, meaning that this function will not return until the feature
        has been enabled.

        """
        logger.info("Enabling reception of response from Device Provisioning Service...")

        subscription_complete = EventedCallback()
        self._provisioning_pipeline.enable_responses(callback=subscription_complete)

        handle_result(subscription_complete)

        logger.info("Successfully subscribed to Device Provisioning Service to receive responses")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module provides a base class for renewable token authentication providers"""

import time
import abc
import logging
import math
import six
import weakref
from threading import Timer
import six.moves.urllib as urllib
from .authentication_provider import AuthenticationProvider

logger = logging.getLogger(__name__)

_device_keyname_token_format = "SharedAccessSignature sr={}&sig={}&se={}&skn={}"
_device_token_format = "SharedAccessSignature sr={}&sig={}&se={}"

# Length of time, in seconds, that a SAS token is valid for.
DEFAULT_TOKEN_VALIDITY_PERIOD = 3600

# Length of time, in seconds, before a token expires that we want to begin renewing it.
DEFAULT_TOKEN_RENEWAL_MARGIN = 120


@six.add_metaclass(abc.ABCMeta)
class BaseRenewableTokenAuthenticationProvider(AuthenticationProvider):
    """A base class for authentication providers which are based on SAS (Shared
    Authentication Signature) strings which are able to be renewed.

    The SAS token string renewal is based on a signing function that is used
    to create the sig field of the SAS string.  This base implements all
    functionality for SAS string creation except for the signing function,
    which is expected to be provided by derived objects.  This base also
    implements the functionality necessary for timing and executing the
    token renewal operation.
    """

    def __init__(self, hostname, device_id, module_id=None):
        """Initializer for Renewable Token Authentication Provider.

        This object is intended as a base class and cannot be used directly.
        A derived class which provides a signing function (such as
        SymmetricKeyAuthenticationProvider or IoTEdgeAuthenticationProvider)
        should be used instead.

        :param str hostname: The hostname
        :param str device_id: The device ID
        :param str module_id: The module ID (optional)
        """

        super(BaseRenewableTokenAuthenticationProvider, self).__init__(
            hostname=hostname, device_id=device_id, module_id=module_id
        )
        self.token_validity_period = DEFAULT_TOKEN_VALIDITY_PERIOD
        self.token_renewal_margin = DEFAULT_TOKEN_RENEWAL_MARGIN
        self._token_update_timer = None
        self.shared_access_key_name = None
        self.sas_token_str = None
        self.on_sas_token_updated_handler_list = []

    def __del__(self):
        self._cancel_token_update_timer()

    def generate_new_sas_token(self):
        """Force the SAS token to update itself.

        This will cause a new sas token to be created using the _sign function.
        This token is valid for roughly self.token_validity_period second.

        This validity period can only be roughly enforced because it relies on the
        coordination of clocks between the client device and the service.  If the two
        different machines have different definitions of "now", most likely because
        of clock drift, then they will also have different notions of when a token will
        expire.  This algorithm atempts to compensate for clock drift by taking
        self.token_renewal_margin into account when deciding when to renew a token.

        If self.token_udpate_callback is set, this callback will be called to notify the
        pipeline that a new token is available.  The pipeline is responsible for doing
        whatever is necessary to leverage the new token when the on_sas_token_updated_handler_list
        function is called.

        The token that is generated expires at some point in the future, based on the token
        renewal interval and the token renewal margin.  When a token is first generated, the
        authorization provider object will set a timer which will be responsible for renewing
        the token before the it expires.  When this timer fires, it will automatically generate
        a new sas token and notify the pipeline by calling self.on_sas_token_updated_handler_list.

        The token update timer is set based on two numbers: self.token_validity_period and
        self.token_renewal_margin

        The first number is the validity period.  This defines the amount of time that the token
        is valid.  The interval is encoded in the token as an offset from the current time,
        as based on the Unix epoch.  In other words, the expiry (se=) value in the token
        is the number of seconds after 00:00 on January 1, 1970 that the token expires.

        The second number that defines the token renewal behavior is the margin.  This is
        the number of seconds before expiration that we want to generate a new token.  Since
        the clocks on different computers can drift over time, they will all have different
        definitions of what "now" is, so the margin needs to be set so there is a
        very small chance that there is no time overlap where one computer thinks the token
        is expired and another doesn't.

        When the timer is set to renew the SAS token, the timer is set for
        (token_validity_period - token_renewal_margin) seconds in the future.  In this way,
        the token will be renewed close to it's expiration time, but not so close that
        we risk a problem caused by clock drift.

        :return: None
        """
        logger.info(
            "Generating new SAS token for (%s,%s) that expires %d seconds in the future",
            self.device_id,
            self.module_id,
            self.token_validity_period,
        )
        expiry = int(math.floor(time.time()) + self.token_validity_period)
        resource_uri = self.hostname + "/devices/" + self.device_id
        if self.module_id:
            resource_uri += "/modules/" + self.module_id
        quoted_resource_uri = urllib.parse.quote_plus(resource_uri)

        signature = self._sign(quoted_resource_uri, expiry)

        if self.shared_access_key_name:
            token = _device_keyname_token_format.format(
                quoted_resource_uri, signature, str(expiry), self.shared_access_key_name
            )
        else:
            token = _device_token_format.format(quoted_resource_uri, signature, str(expiry))

        self.sas_token_str = str(token)
        self._schedule_token_update(self.token_validity_period - self.token_renewal_margin)
        self._notify_token_updated()

    def _cancel_token_update_timer(self):
        """Cancel any future token update operations.  This is typically done as part of a
        teardown operation.
        """
        t = self._token_update_timer
        self._token_update_timer = None
        if t:
            logger.debug(
                "Canceling token update timer for (%s,%s)",
                self.device_id,
                self.module_id if self.module_id else "",
            )
            t.cancel()

    def _schedule_token_update(self, seconds_until_update):
        """Schedule an automatic sas token update to take place seconds_until_update seconds in
        the future.  If an update was previously scheduled, this method shall cancel the
        previously-scheduled update and schedule a new update.
        """
        self._cancel_token_update_timer()
        logger.debug(
            "Scheduling token update for (%s,%s) for %d seconds in the future",
            self.device_id,
            self.module_id,
            seconds_until_update,
        )

        # It's important to use a weak reference to self inside this timer function
        # because we don't want the timer to prevent this object (`self`) from being collected.
        #
        # We want `self` to get collected when the pipeline gets collected, and
        # we want the pipeline to get collected when the client object gets collected.
        # This way, everything gets cleaned up when the user is done with the client object,
        # as expected.
        #
        # If timerfunc used `self` directly, that would be a strong reference, and that strong
        # reference would prevent `self` from being collected as long as the timer existed.
        #
        # If this isn't collected when the client is collected, then the object that implements the
        # on_sas_token_updated_hndler doesn't get collected.  Since that object is part of the
        # pipeline, a major part of the pipeline ends up staying around, probably orphaned from
        # the client.  Since that orphaned part of the pipeline contains Paho, bad things can happen
        # if we don't clean up Paho correctly.  This is especially noticable if one process
        # destroys a client object and creates a new one.
        #
        self_weakref = weakref.ref(self)

        def timerfunc():
            this = self_weakref()
            logger.debug("Timed SAS update for (%s,%s)", this.device_id, this.module_id)
            this.generate_new_sas_token()

        self._token_update_timer = Timer(seconds_until_update, timerfunc)
        self._token_update_timer.daemon = True
        self._token_update_timer.start()

    def _notify_token_updated(self):
        """Notify clients that the SAS token has been updated by calling self.on_sas_token_updated.
        In response to this event, clients should re-initiate their connection in order to use
        the updated sas token.
        """
        if bool(len(self.on_sas_token_updated_handler_list)):
            logger.debug(
                "sending token update notification for (%s, %s)", self.device_id, self.module_id
            )
            for x in self.on_sas_token_updated_handler_list:
                x()
        else:
            logger.warning(
                "_notify_token_updated: on_sas_token_updated_handler_list not set.  Doing nothing."
            )

    def get_current_sas_token(self):
        """Get the current SharedAuthenticationSignature string.

        This string can be used to authenticate with an Azure IoT Hub or Azure IoT Edge Hub service.

        If a SAS token has not yet been created yet, this function call the generate_new_sas_token
        function to create a new token and schedule the update timer.  See the documentation for
        generate_new_sas_token for more detail.

        :return: The current shared access signature token in string form.
        """
        if not self.sas_token_str:
            self.generate_new_sas_token()
        return self.sas_token_str

    @abc.abstractmethod
    def _sign(self, quoted_resource_uri, expiry):
        """Create and return a new signature for this object.  The caller is responsible
        for placing the signature inside the sig field of a SAS token string.
        """
        pass



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from . import PipelineOperation


class SetHTTPConnectionArgsOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to connect to a server using the HTTP protocol.

    This operation is in the group of HTTP operations because its attributes are very specific to the HTTP protocol.
    """

    def __init__(
        self, hostname, callback, server_verification_cert=None, client_cert=None, sas_token=None
    ):
        """
        Initializer for SetHTTPConnectionArgsOperation objects.
        :param str hostname: The hostname of the HTTP server we will eventually connect to
        :param str server_verification_cert: (Optional) The server verification certificate to use
            if the HTTP server that we're going to connect to uses server-side TLS
        :param X509 client_cert: (Optional) The x509 object containing a client certificate and key used to connect
            to the HTTP service
        :param str sas_token: The token string which will be used to authenticate with the service
        :param Function callback: The function that gets called when this operation is complete or has failed.
            The callback function must accept A PipelineOperation object which indicates the specific operation which
            has completed or failed.
        """
        super(SetHTTPConnectionArgsOperation, self).__init__(callback=callback)
        self.hostname = hostname
        self.server_verification_cert = server_verification_cert
        self.client_cert = client_cert
        self.sas_token = sas_token


class HTTPRequestAndResponseOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to connect to a server using the HTTP protocol.

    This operation is in the group of HTTP operations because its attributes are very specific to the HTTP protocol.
    """

    def __init__(self, method, path, headers, body, query_params, callback):
        """
        Initializer for HTTPPublishOperation objects.
        :param str method: The HTTP method used in the request
        :param str path: The path to be used in the request url
        :param dict headers: The headers to be used in the HTTP request
        :param str body: The body to be provided with the HTTP request
        :param str query_params: The query parameters to be used in the request url
        :param Function callback: The function that gets called when this operation is complete or has failed.
            The callback function must accept A PipelineOperation object which indicates the specific operation which
            has completed or failed.
        """
        super(HTTPRequestAndResponseOperation, self).__init__(callback=callback)
        self.method = method
        self.path = path
        self.headers = headers
        self.body = body
        self.query_params = query_params
        self.status_code = None
        self.response_body = None
        self.reason = None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import platform
from azure.iot.device.constant import VERSION, IOTHUB_IDENTIFIER, PROVISIONING_IDENTIFIER

python_runtime = platform.python_version()
os_type = platform.system()
os_release = platform.version()
architecture = platform.machine()


class ProductInfo(object):
    """
    A class for creating product identifiers or agent strings for IotHub as well as Provisioning.
    """

    @staticmethod
    def _get_common_user_agent():
        return "({python_runtime};{os_type} {os_release};{architecture})".format(
            python_runtime=python_runtime,
            os_type=os_type,
            os_release=os_release,
            architecture=architecture,
        )

    @staticmethod
    def get_iothub_user_agent():
        """
        Create the user agent for IotHub
        """
        return "{iothub_iden}/{version}{common}".format(
            iothub_iden=IOTHUB_IDENTIFIER,
            version=VERSION,
            common=ProductInfo._get_common_user_agent(),
        )

    @staticmethod
    def get_provisioning_user_agent():
        """
        Create the user agent for Provisioning
        """
        return "{provisioning_iden}/{version}{common}".format(
            provisioning_iden=PROVISIONING_IDENTIFIER,
            version=VERSION,
            common=ProductInfo._get_common_user_agent(),
        )



"""Azure IoT Hub Device SDK Authentication

This package provides authentication-related functionality for use with the
Azure IoT Hub Device SDK.
"""

from .sk_authentication_provider import SymmetricKeyAuthenticationProvider
from .sas_authentication_provider import SharedAccessSignatureAuthenticationProvider
from .iotedge_authentication_provider import IoTEdgeAuthenticationProvider, IoTEdgeError
from .x509_authentication_provider import X509AuthenticationProvider

__all__ = [
    "SymmetricKeyAuthenticationProvider",
    "SharedAccessSignatureAuthenticationProvider",
    "IoTEdgeAuthenticationProvider",
    "IoTEdgeError",
    "X509AuthenticationProvider",
]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from azure.iot.device.common.pipeline import pipeline_ops_base, pipeline_thread
from azure.iot.device.common.pipeline.pipeline_stages_base import PipelineStage
from . import pipeline_ops_provisioning
from azure.iot.device import exceptions
from azure.iot.device.provisioning.pipeline import constant
from azure.iot.device.provisioning.models.registration_result import (
    RegistrationResult,
    RegistrationState,
)
import logging
import weakref
import json
from threading import Timer
import time

logger = logging.getLogger(__name__)


class UseSecurityClientStage(PipelineStage):
    """
    PipelineStage which extracts relevant SecurityClient values for a new
    SetProvisioningClientConnectionArgsOperation.

    All other operations are passed down.
    """

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_provisioning.SetSymmetricKeySecurityClientOperation):

            security_client = op.security_client
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation,
                provisioning_host=security_client.provisioning_host,
                registration_id=security_client.registration_id,
                id_scope=security_client.id_scope,
                sas_token=security_client.get_current_sas_token(),
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_provisioning.SetX509SecurityClientOperation):
            security_client = op.security_client
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation,
                provisioning_host=security_client.provisioning_host,
                registration_id=security_client.registration_id,
                id_scope=security_client.id_scope,
                client_cert=security_client.get_x509_certificate(),
            )
            self.send_op_down(worker_op)

        else:
            super(UseSecurityClientStage, self)._run_op(op)


class CommonProvisioningStage(PipelineStage):
    """
    This is a super stage that the RegistrationStage and PollingStatusStage of
    provisioning would both use. It contains some common functions like decoding response
    and retrieving error, retrieving registration status, retrieving operation id
    and forming a complete result.
    """

    @pipeline_thread.runs_on_pipeline_thread
    def _clear_timeout_timer(self, op, error):
        """
        Clearing timer for provisioning operations (Register and PollStatus)
        when they respond back from service.
        """
        if op.provisioning_timeout_timer:
            logger.debug("{}({}): Cancelling provisioning timeout timer".format(self.name, op.name))
            op.provisioning_timeout_timer.cancel()
            op.provisioning_timeout_timer = None

    @staticmethod
    def _decode_response(provisioning_op):
        return json.loads(provisioning_op.response_body.decode("utf-8"))

    @staticmethod
    def _form_complete_result(operation_id, decoded_response, status):
        """
        Create the registration result from the complete decoded json response for details regarding the registration process.
        """
        decoded_state = decoded_response.get("registrationState", None)
        registration_state = None
        if decoded_state is not None:
            registration_state = RegistrationState(
                device_id=decoded_state.get("deviceId", None),
                assigned_hub=decoded_state.get("assignedHub", None),
                sub_status=decoded_state.get("substatus", None),
                created_date_time=decoded_state.get("createdDateTimeUtc", None),
                last_update_date_time=decoded_state.get("lastUpdatedDateTimeUtc", None),
                etag=decoded_state.get("etag", None),
                payload=decoded_state.get("payload", None),
            )

        registration_result = RegistrationResult(
            operation_id=operation_id, status=status, registration_state=registration_state
        )
        return registration_result

    def _process_service_error_status_code(self, original_provisioning_op, request_response_op):
        logger.error(
            "{stage_name}({op_name}): Received error with status code {status_code} for {prov_op_name} request operation".format(
                stage_name=self.name,
                op_name=request_response_op.name,
                prov_op_name=request_response_op.request_type,
                status_code=request_response_op.status_code,
            )
        )
        logger.error(
            "{stage_name}({op_name}): Response body: {body}".format(
                stage_name=self.name,
                op_name=request_response_op.name,
                body=request_response_op.response_body,
            )
        )
        original_provisioning_op.complete(
            error=exceptions.ServiceError(
                "{prov_op_name} request returned a service error status code {status_code}".format(
                    prov_op_name=request_response_op.request_type,
                    status_code=request_response_op.status_code,
                )
            )
        )

    def _process_retry_status_code(self, error, original_provisioning_op, request_response_op):
        retry_interval = (
            int(request_response_op.retry_after, 10)
            if request_response_op.retry_after is not None
            else constant.DEFAULT_POLLING_INTERVAL
        )

        self_weakref = weakref.ref(self)

        @pipeline_thread.invoke_on_pipeline_thread_nowait
        def do_retry_after():
            this = self_weakref()
            logger.info(
                "{stage_name}({op_name}): retrying".format(
                    stage_name=this.name, op_name=request_response_op.name
                )
            )
            original_provisioning_op.retry_after_timer.cancel()
            original_provisioning_op.retry_after_timer = None
            original_provisioning_op.completed = False
            this.run_op(original_provisioning_op)

        logger.warning(
            "{stage_name}({op_name}): Op needs retry with interval {interval} because of {error}. Setting timer.".format(
                stage_name=self.name,
                op_name=request_response_op.name,
                interval=retry_interval,
                error=error,
            )
        )

        logger.debug("{}({}): Creating retry timer".format(self.name, request_response_op.name))
        original_provisioning_op.retry_after_timer = Timer(retry_interval, do_retry_after)
        original_provisioning_op.retry_after_timer.start()

    @staticmethod
    def _process_failed_and_assigned_registration_status(
        error,
        operation_id,
        decoded_response,
        registration_status,
        original_provisioning_op,
        request_response_op,
    ):
        complete_registration_result = CommonProvisioningStage._form_complete_result(
            operation_id=operation_id, decoded_response=decoded_response, status=registration_status
        )
        original_provisioning_op.registration_result = complete_registration_result
        if registration_status == "failed":
            error = exceptions.ServiceError(
                "Query Status operation returned a failed registration status  with a status code of {status_code}".format(
                    status_code=request_response_op.status_code
                )
            )
        original_provisioning_op.complete(error=error)

    @staticmethod
    def _process_unknown_registration_status(
        registration_status, original_provisioning_op, request_response_op
    ):
        error = exceptions.ServiceError(
            "Query Status Operation encountered an invalid registration status {status} with a status code of {status_code}".format(
                status=registration_status, status_code=request_response_op.status_code
            )
        )
        original_provisioning_op.complete(error=error)


class PollingStatusStage(CommonProvisioningStage):
    """
    This stage is responsible for sending the query request once initial response
    is received from the registration response.
    Upon the receipt of the response this stage decides whether
    to send another query request or complete the procedure.
    """

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_provisioning.PollStatusOperation):
            query_status_op = op
            self_weakref = weakref.ref(self)

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def query_timeout():
                this = self_weakref()
                logger.info(
                    "{stage_name}({op_name}): returning timeout error".format(
                        stage_name=this.name, op_name=op.name
                    )
                )
                query_status_op.complete(
                    error=(
                        exceptions.ServiceError(
                            "Operation timed out before provisioning service could respond for {op_type} operation".format(
                                op_type=constant.QUERY
                            )
                        )
                    )
                )

            logger.debug("{}({}): Creating provisioning timeout timer".format(self.name, op.name))
            query_status_op.provisioning_timeout_timer = Timer(
                constant.DEFAULT_TIMEOUT_INTERVAL, query_timeout
            )
            query_status_op.provisioning_timeout_timer.start()

            def on_query_response(op, error):
                self._clear_timeout_timer(query_status_op, error)
                logger.debug(
                    "{stage_name}({op_name}): Received response with status code {status_code} for PollStatusOperation with operation id {oper_id}".format(
                        stage_name=self.name,
                        op_name=op.name,
                        status_code=op.status_code,
                        oper_id=op.query_params["operation_id"],
                    )
                )

                if error:
                    logger.error(
                        "{stage_name}({op_name}): Received error for {prov_op_name} operation".format(
                            stage_name=self.name, op_name=op.name, prov_op_name=op.request_type
                        )
                    )
                    query_status_op.complete(error=error)

                else:
                    if 300 <= op.status_code < 429:
                        self._process_service_error_status_code(query_status_op, op)

                    elif op.status_code >= 429:
                        self._process_retry_status_code(error, query_status_op, op)

                    else:
                        decoded_response = self._decode_response(op)
                        operation_id = decoded_response.get("operationId", None)
                        registration_status = decoded_response.get("status", None)
                        if registration_status == "assigning":
                            polling_interval = (
                                int(op.retry_after, 10)
                                if op.retry_after is not None
                                else constant.DEFAULT_POLLING_INTERVAL
                            )
                            self_weakref = weakref.ref(self)

                            @pipeline_thread.invoke_on_pipeline_thread_nowait
                            def do_polling():
                                this = self_weakref()
                                logger.info(
                                    "{stage_name}({op_name}): retrying".format(
                                        stage_name=this.name, op_name=op.name
                                    )
                                )
                                query_status_op.polling_timer.cancel()
                                query_status_op.polling_timer = None
                                query_status_op.completed = False
                                this.run_op(query_status_op)

                            logger.info(
                                "{stage_name}({op_name}): Op needs retry with interval {interval} because of {error}. Setting timer.".format(
                                    stage_name=self.name,
                                    op_name=op.name,
                                    interval=polling_interval,
                                    error=error,
                                )
                            )

                            logger.debug(
                                "{}({}): Creating polling timer".format(self.name, op.name)
                            )
                            query_status_op.polling_timer = Timer(polling_interval, do_polling)
                            query_status_op.polling_timer.start()

                        elif registration_status == "assigned" or registration_status == "failed":
                            self._process_failed_and_assigned_registration_status(
                                error=error,
                                operation_id=operation_id,
                                decoded_response=decoded_response,
                                registration_status=registration_status,
                                original_provisioning_op=query_status_op,
                                request_response_op=op,
                            )

                        else:
                            self._process_unknown_registration_status(
                                registration_status=registration_status,
                                original_provisioning_op=query_status_op,
                                request_response_op=op,
                            )

            self.send_op_down(
                pipeline_ops_base.RequestAndResponseOperation(
                    request_type=constant.QUERY,
                    method="GET",
                    resource_location="/",
                    query_params={"operation_id": query_status_op.operation_id},
                    request_body=query_status_op.request_payload,
                    callback=on_query_response,
                )
            )

        else:
            super(PollingStatusStage, self)._run_op(op)


class RegistrationStage(CommonProvisioningStage):
    """
    This is the first stage that decides converts a registration request
    into a normal request and response operation.
    Upon the receipt of the response this stage decides whether
    to send another registration request or send a query request.
    Depending on the status and result of the response
    this stage may also complete the registration process.
    """

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_provisioning.RegisterOperation):
            initial_register_op = op
            self_weakref = weakref.ref(self)

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def register_timeout():
                this = self_weakref()
                logger.info(
                    "{stage_name}({op_name}): returning timeout error".format(
                        stage_name=this.name, op_name=op.name
                    )
                )
                initial_register_op.complete(
                    error=(
                        exceptions.ServiceError(
                            "Operation timed out before provisioning service could respond for {op_type} operation".format(
                                op_type=constant.REGISTER
                            )
                        )
                    )
                )

            logger.debug("{}({}): Creating provisioning timeout timer".format(self.name, op.name))
            initial_register_op.provisioning_timeout_timer = Timer(
                constant.DEFAULT_TIMEOUT_INTERVAL, register_timeout
            )
            initial_register_op.provisioning_timeout_timer.start()

            def on_registration_response(op, error):
                self._clear_timeout_timer(initial_register_op, error)
                logger.debug(
                    "{stage_name}({op_name}): Received response with status code {status_code} for RegisterOperation".format(
                        stage_name=self.name, op_name=op.name, status_code=op.status_code
                    )
                )
                if error:
                    logger.error(
                        "{stage_name}({op_name}): Received error for {prov_op_name} operation".format(
                            stage_name=self.name, op_name=op.name, prov_op_name=op.request_type
                        )
                    )
                    initial_register_op.complete(error=error)

                else:

                    if 300 <= op.status_code < 429:
                        self._process_service_error_status_code(initial_register_op, op)

                    elif op.status_code >= 429:
                        self._process_retry_status_code(error, initial_register_op, op)

                    else:
                        decoded_response = self._decode_response(op)
                        operation_id = decoded_response.get("operationId", None)
                        registration_status = decoded_response.get("status", None)

                        if registration_status == "assigning":
                            self_weakref = weakref.ref(self)

                            def copy_result_to_original_op(op, error):
                                logger.debug(
                                    "Copying registration result from Query Status Op to Registration Op"
                                )
                                initial_register_op.registration_result = op.registration_result
                                initial_register_op.error = error

                            @pipeline_thread.invoke_on_pipeline_thread_nowait
                            def do_query_after_interval():
                                this = self_weakref()
                                initial_register_op.polling_timer.cancel()
                                initial_register_op.polling_timer = None

                                logger.info(
                                    "{stage_name}({op_name}): polling".format(
                                        stage_name=this.name, op_name=op.name
                                    )
                                )

                                query_worker_op = initial_register_op.spawn_worker_op(
                                    worker_op_type=pipeline_ops_provisioning.PollStatusOperation,
                                    request_payload=" ",
                                    operation_id=operation_id,
                                    callback=copy_result_to_original_op,
                                )

                                self.send_op_down(query_worker_op)

                            logger.warning(
                                "{stage_name}({op_name}): Op will transition into polling after interval {interval}.  Setting timer.".format(
                                    stage_name=self.name,
                                    op_name=op.name,
                                    interval=constant.DEFAULT_POLLING_INTERVAL,
                                )
                            )

                            logger.debug(
                                "{}({}): Creating polling timer".format(self.name, op.name)
                            )
                            initial_register_op.polling_timer = Timer(
                                constant.DEFAULT_POLLING_INTERVAL, do_query_after_interval
                            )
                            initial_register_op.polling_timer.start()

                        elif registration_status == "failed" or registration_status == "assigned":
                            self._process_failed_and_assigned_registration_status(
                                error=error,
                                operation_id=operation_id,
                                decoded_response=decoded_response,
                                registration_status=registration_status,
                                original_provisioning_op=initial_register_op,
                                request_response_op=op,
                            )

                        else:
                            self._process_unknown_registration_status(
                                registration_status=registration_status,
                                original_provisioning_op=initial_register_op,
                                request_response_op=op,
                            )

            registration_payload = DeviceRegistrationPayload(
                registration_id=initial_register_op.registration_id,
                custom_payload=initial_register_op.request_payload,
            )
            self.send_op_down(
                pipeline_ops_base.RequestAndResponseOperation(
                    request_type=constant.REGISTER,
                    method="PUT",
                    resource_location="/",
                    request_body=registration_payload.get_json_string(),
                    callback=on_registration_response,
                )
            )

        else:
            super(RegistrationStage, self)._run_op(op)


class DeviceRegistrationPayload(object):
    """
    The class representing the payload that needs to be sent to the service.
    """

    def __init__(self, registration_id, custom_payload=None):
        # This is not a convention to name variables in python but the
        # DPS service spec needs the name to be exact for it to work
        self.registrationId = registration_id
        self.payload = custom_payload

    def get_json_string(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import json
from six.moves import urllib
from azure.iot.device.common import version_compat
from azure.iot.device.common.pipeline import (
    pipeline_events_base,
    pipeline_ops_base,
    pipeline_ops_mqtt,
    pipeline_events_mqtt,
    PipelineStage,
    pipeline_thread,
)
from azure.iot.device.iothub.models import Message, MethodRequest
from . import pipeline_ops_iothub, pipeline_events_iothub, mqtt_topic_iothub
from . import constant as pipeline_constant
from . import exceptions as pipeline_exceptions
from azure.iot.device import constant as pkg_constant
from azure.iot.device.product_info import ProductInfo

logger = logging.getLogger(__name__)


class IoTHubMQTTTranslationStage(PipelineStage):
    """
    PipelineStage which converts other Iot and IoTHub operations into MQTT operations.  This stage also
    converts mqtt pipeline events into Iot and IoTHub pipeline events.
    """

    def __init__(self):
        super(IoTHubMQTTTranslationStage, self).__init__()
        self.feature_to_topic = {}
        self.device_id = None
        self.module_id = None

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):

        if isinstance(op, pipeline_ops_iothub.SetIoTHubConnectionArgsOperation):
            self.device_id = op.device_id
            self.module_id = op.module_id

            # if we get auth provider args from above, we save some, use some to build topic names,
            # and always pass it down because we know that the MQTT protocol stage will also want
            # to receive these args.
            self._set_topic_names(device_id=op.device_id, module_id=op.module_id)

            if op.module_id:
                client_id = "{}/{}".format(op.device_id, op.module_id)
            else:
                client_id = op.device_id

            # For MQTT, the entire user agent string should be appended to the username field in the connect packet
            # For example, the username may look like this without custom parameters:
            # yosephsandboxhub.azure-devices.net/alpha/?api-version=2018-06-30&DeviceClientType=py-azure-iot-device%2F2.0.0-preview.12
            # The customer user agent string would simply be appended to the end of this username, in URL Encoded format.
            query_param_seq = [
                ("api-version", pkg_constant.IOTHUB_API_VERSION),
                (
                    "DeviceClientType",
                    ProductInfo.get_iothub_user_agent()
                    + str(self.pipeline_root.pipeline_configuration.product_info),
                ),
            ]
            username = "{hostname}/{client_id}/?{query_params}".format(
                hostname=op.hostname,
                client_id=client_id,
                query_params=version_compat.urlencode(
                    query_param_seq, quote_via=urllib.parse.quote
                ),
            )

            if op.gateway_hostname:
                hostname = op.gateway_hostname
            else:
                hostname = op.hostname

            # TODO: test to make sure client_cert and sas_token travel down correctly
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.SetMQTTConnectionArgsOperation,
                client_id=client_id,
                hostname=hostname,
                username=username,
                server_verification_cert=op.server_verification_cert,
                client_cert=op.client_cert,
                sas_token=op.sas_token,
            )
            self.send_op_down(worker_op)

        elif (
            isinstance(op, pipeline_ops_base.UpdateSasTokenOperation)
            and self.pipeline_root.connected
        ):
            logger.debug(
                "{}({}): Connected.  Passing op down and reauthorizing after token is updated.".format(
                    self.name, op.name
                )
            )

            # make a callback that either fails the UpdateSasTokenOperation (if the lower level failed it),
            # or issues a ReauthorizeConnectionOperation (if the lower level returned success for the UpdateSasTokenOperation)
            def on_token_update_complete(op, error):
                if error:
                    logger.error(
                        "{}({}) token update failed.  returning failure {}".format(
                            self.name, op.name, error
                        )
                    )
                else:
                    logger.debug(
                        "{}({}) token update succeeded.  reauthorizing".format(self.name, op.name)
                    )

                    # Stop completion of Token Update op, and only continue upon completion of ReauthorizeConnectionOperation
                    op.halt_completion()
                    worker_op = op.spawn_worker_op(
                        worker_op_type=pipeline_ops_base.ReauthorizeConnectionOperation
                    )

                    self.send_op_down(worker_op)

            # now, pass the UpdateSasTokenOperation down with our new callback.
            op.add_callback(on_token_update_complete)
            self.send_op_down(op)

        elif isinstance(op, pipeline_ops_iothub.SendD2CMessageOperation) or isinstance(
            op, pipeline_ops_iothub.SendOutputEventOperation
        ):
            # Convert SendTelementry and SendOutputEventOperation operations into MQTT Publish operations
            topic = mqtt_topic_iothub.encode_message_properties_in_topic(
                op.message, self.telemetry_topic
            )
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.MQTTPublishOperation,
                topic=topic,
                payload=op.message.data,
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_iothub.SendMethodResponseOperation):
            # Sending a Method Response gets translated into an MQTT Publish operation
            topic = mqtt_topic_iothub.get_method_topic_for_publish(
                op.method_response.request_id, str(op.method_response.status)
            )
            payload = json.dumps(op.method_response.payload)
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.MQTTPublishOperation, topic=topic, payload=payload
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_base.EnableFeatureOperation):
            # Enabling a feature gets translated into an MQTT subscribe operation
            topic = self.feature_to_topic[op.feature_name]
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.MQTTSubscribeOperation, topic=topic
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_base.DisableFeatureOperation):
            # Disabling a feature gets turned into an MQTT unsubscribe operation
            topic = self.feature_to_topic[op.feature_name]
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.MQTTUnsubscribeOperation, topic=topic
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_base.RequestOperation):
            if op.request_type == pipeline_constant.TWIN:
                topic = mqtt_topic_iothub.get_twin_topic_for_publish(
                    method=op.method,
                    resource_location=op.resource_location,
                    request_id=op.request_id,
                )
                worker_op = op.spawn_worker_op(
                    worker_op_type=pipeline_ops_mqtt.MQTTPublishOperation,
                    topic=topic,
                    payload=op.request_body,
                )
                self.send_op_down(worker_op)
            else:
                raise pipeline_exceptions.OperationError(
                    "RequestOperation request_type {} not supported".format(op.request_type)
                )

        else:
            # All other operations get passed down
            super(IoTHubMQTTTranslationStage, self)._run_op(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _set_topic_names(self, device_id, module_id):
        """
        Build topic names based on the device_id and module_id passed.
        """
        self.telemetry_topic = mqtt_topic_iothub.get_telemetry_topic_for_publish(
            device_id, module_id
        )
        self.feature_to_topic = {
            pipeline_constant.C2D_MSG: (mqtt_topic_iothub.get_c2d_topic_for_subscribe(device_id)),
            pipeline_constant.INPUT_MSG: (
                mqtt_topic_iothub.get_input_topic_for_subscribe(device_id, module_id)
            ),
            pipeline_constant.METHODS: (mqtt_topic_iothub.get_method_topic_for_subscribe()),
            pipeline_constant.TWIN: (mqtt_topic_iothub.get_twin_response_topic_for_subscribe()),
            pipeline_constant.TWIN_PATCHES: (
                mqtt_topic_iothub.get_twin_patch_topic_for_subscribe()
            ),
        }

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        """
        Pipeline Event handler function to convert incoming MQTT messages into the appropriate IoTHub
        events, based on the topic of the message
        """
        if isinstance(event, pipeline_events_mqtt.IncomingMQTTMessageEvent):
            topic = event.topic

            if mqtt_topic_iothub.is_c2d_topic(topic, self.device_id):
                message = Message(event.payload)
                mqtt_topic_iothub.extract_message_properties_from_topic(topic, message)
                self.send_event_up(pipeline_events_iothub.C2DMessageEvent(message))

            elif mqtt_topic_iothub.is_input_topic(topic, self.device_id, self.module_id):
                message = Message(event.payload)
                mqtt_topic_iothub.extract_message_properties_from_topic(topic, message)
                input_name = mqtt_topic_iothub.get_input_name_from_topic(topic)
                self.send_event_up(pipeline_events_iothub.InputMessageEvent(input_name, message))

            elif mqtt_topic_iothub.is_method_topic(topic):
                request_id = mqtt_topic_iothub.get_method_request_id_from_topic(topic)
                method_name = mqtt_topic_iothub.get_method_name_from_topic(topic)
                method_received = MethodRequest(
                    request_id=request_id,
                    name=method_name,
                    payload=json.loads(event.payload.decode("utf-8")),
                )
                self.send_event_up(pipeline_events_iothub.MethodRequestEvent(method_received))

            elif mqtt_topic_iothub.is_twin_response_topic(topic):
                request_id = mqtt_topic_iothub.get_twin_request_id_from_topic(topic)
                status_code = int(mqtt_topic_iothub.get_twin_status_code_from_topic(topic))
                self.send_event_up(
                    pipeline_events_base.ResponseEvent(
                        request_id=request_id, status_code=status_code, response_body=event.payload
                    )
                )

            elif mqtt_topic_iothub.is_twin_desired_property_patch_topic(topic):
                self.send_event_up(
                    pipeline_events_iothub.TwinDesiredPropertiesPatchEvent(
                        patch=json.loads(event.payload.decode("utf-8"))
                    )
                )

            else:
                logger.debug("Unknown topic: {} passing up to next handler".format(topic))
                self.send_event_up(event)

        else:
            # all other messages get passed up
            super(IoTHubMQTTTranslationStage, self)._handle_pipeline_event(event)



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module defines functions intended for providing compatibility between
different versions of Python"""

from six.moves import urllib


def urlencode(query, quote_via=urllib.parse.quote_plus, safe=""):
    """ Custom implementation of urllib.parse.urlencode().

    This is necessary because prior to Python 3.5, urlencode() always encodes via
    quote_plus() rather than quote(). This is generally not desirable for MQTT, as
    it will translate ' ' into '+' rather than '%20', and '+' is not allowed in the
    topic strings for MQTT publish.

    Starting in 3.5, the included library function for urlencode() allows you to specify
    which style of encoding you want, however this feature is not available in 2.7 and so
    we must implement it ourselves.
    """
    if isinstance(query, list):
        encoded = "&".join(
            ["{}={}".format(quote_via(k, safe=safe), quote_via(v, safe=safe)) for (k, v) in query]
        )
    elif isinstance(query, dict):
        encoded = "&".join(
            [
                "{}={}".format(quote_via(k, safe=safe), quote_via(v, safe=safe))
                for k, v in query.items()
            ]
        )
    else:
        raise TypeError("Invalid type for 'query'")
    return encoded



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import base64
import hmac
import hashlib
import logging
import six.moves.urllib as urllib
from .base_renewable_token_authentication_provider import BaseRenewableTokenAuthenticationProvider

logger = logging.getLogger(__name__)

DELIMITER = ";"
VALUE_SEPARATOR = "="

HOST_NAME = "HostName"
SHARED_ACCESS_KEY_NAME = "SharedAccessKeyName"
SHARED_ACCESS_KEY = "SharedAccessKey"
SHARED_ACCESS_SIGNATURE = "SharedAccessSignature"
DEVICE_ID = "DeviceId"
MODULE_ID = "ModuleId"
GATEWAY_HOST_NAME = "GatewayHostName"

_valid_keys = [
    HOST_NAME,
    SHARED_ACCESS_KEY_NAME,
    SHARED_ACCESS_KEY,
    SHARED_ACCESS_SIGNATURE,
    DEVICE_ID,
    MODULE_ID,
    GATEWAY_HOST_NAME,
]


class SymmetricKeyAuthenticationProvider(BaseRenewableTokenAuthenticationProvider):
    """
    A Symmetric Key Authentication Provider. This provider needs to create the i
    Shared Access Signature that would be needed to connect to the IoT Hub.
    """

    def __init__(
        self,
        hostname,
        device_id,
        module_id,
        shared_access_key,
        shared_access_key_name=None,
        gateway_hostname=None,
    ):
        """

        Constructor for SymmetricKey Authentication Provider
        """
        logger.info(
            "Using Shared Key authentication for {%s, %s, %s}", hostname, device_id, module_id
        )

        super(SymmetricKeyAuthenticationProvider, self).__init__(
            hostname=hostname, device_id=device_id, module_id=module_id
        )
        self.shared_access_key = shared_access_key
        self.shared_access_key_name = shared_access_key_name
        self.gateway_hostname = gateway_hostname
        self.server_verification_cert = None

    @staticmethod
    def parse(connection_string):
        """
        This method creates a Symmetric Key Authentication Provider from a given connection string, and sets properties for each of the parsed
        fields in the string. Also validates the required properties of the connection string.
        :param connection_string: The semicolon-delimited string of 'name=value' pairs.
        The input may look like the following formations:-
        HostName=<hostname>;DeviceId=<device_id>;SharedAccessKey=<shared_access_key>
        HostName=<hostname>;DeviceId=<device_id>;SharedAccessKeyName=<shared_access_key_name>;SharedAccessKey=<shared_access_key>
        HostName=<hostname>;DeviceId=<device_id>;ModuleId=<module_id>;SharedAccessKey=<shared_access_key>
        :return: The Symmetric Key Authentication Provider constructed
        """
        try:
            cs_args = connection_string.split(DELIMITER)
            d = dict(arg.split(VALUE_SEPARATOR, 1) for arg in cs_args)
        except (ValueError, AttributeError):
            raise ValueError(
                "Connection string is required and should not be empty or blank and must be supplied as a string"
            )
        if len(cs_args) != len(d):
            raise ValueError("Invalid Connection String - Unable to parse")
        if not all(key in _valid_keys for key in d.keys()):
            raise ValueError("Invalid Connection String - Invalid Key")

        _validate_keys(d)

        return SymmetricKeyAuthenticationProvider(
            d.get(HOST_NAME),
            d.get(DEVICE_ID),
            d.get(MODULE_ID),
            d.get(SHARED_ACCESS_KEY),
            d.get(SHARED_ACCESS_KEY_NAME),
            d.get(GATEWAY_HOST_NAME),
        )

    def _sign(self, quoted_resource_uri, expiry):
        """
        Creates the base64-encoded HMAC-SHA256 hash of the string to sign. The string to sign is constructed from the
        resource_uri and expiry and the signing key is constructed from the device_key.
        :param  quoted_resource_uri: the resource URI to encode into the token, already URI-encoded
        :param expiry: an integer value representing the number of seconds since the epoch 00:00:00 UTC on 1 January 1970 at which the token will expire.
        :return: The signature portion of the Sas Token.
        """
        try:
            message = (quoted_resource_uri + "\n" + str(expiry)).encode("utf-8")
            signing_key = base64.b64decode(self.shared_access_key.encode("utf-8"))
            signed_hmac = hmac.HMAC(signing_key, message, hashlib.sha256)
            signature = urllib.parse.quote(base64.b64encode(signed_hmac.digest()))
        except (TypeError, base64.binascii.Error):
            raise ValueError("Unable to build shared access signature from given values")
        return signature


def _validate_keys(d):
    """Raise ValueError if incorrect combination of keys
    """
    host_name = d.get(HOST_NAME)
    shared_access_key_name = d.get(SHARED_ACCESS_KEY_NAME)
    shared_access_key = d.get(SHARED_ACCESS_KEY)
    device_id = d.get(DEVICE_ID)

    if host_name and device_id and shared_access_key:
        pass
    elif host_name and shared_access_key and shared_access_key_name:
        pass
    else:
        raise ValueError("Invalid Connection String - Incomplete")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import base64
import json
import six.moves.urllib as urllib
import requests
import requests_unixsocket
import logging
from .base_renewable_token_authentication_provider import BaseRenewableTokenAuthenticationProvider
from azure.iot.device.common.chainable_exception import ChainableException
from azure.iot.device.product_info import ProductInfo

requests_unixsocket.monkeypatch()

logger = logging.getLogger(__name__)


class IoTEdgeError(ChainableException):
    pass


class IoTEdgeAuthenticationProvider(BaseRenewableTokenAuthenticationProvider):
    """An Azure IoT Edge Authentication Provider.

    This provider creates the Shared Access Signature that would be needed to connenct to the IoT Edge runtime
    """

    def __init__(
        self,
        hostname,
        device_id,
        module_id,
        gateway_hostname,
        module_generation_id,
        workload_uri,
        api_version,
    ):
        """
        Constructor for IoT Edge Authentication Provider
        """

        logger.info("Using IoTEdge authentication for {%s, %s, %s}", hostname, device_id, module_id)

        super(IoTEdgeAuthenticationProvider, self).__init__(
            hostname=hostname, device_id=device_id, module_id=module_id
        )

        self.hsm = IoTEdgeHsm(
            module_id=module_id,
            api_version=api_version,
            module_generation_id=module_generation_id,
            workload_uri=workload_uri,
        )
        self.gateway_hostname = gateway_hostname
        self.server_verification_cert = self.hsm.get_trust_bundle()

    # TODO: reconsider this design when refactoring the BaseRenewableToken auth parent
    # TODO: Consider handling the quoting within this function, and renaming quoted_resource_uri to resource_uri
    def _sign(self, quoted_resource_uri, expiry):
        """
        Creates the signature to be inserted in the SAS token
        :param resource_uri: the resource URI to encode into the token
        :param expiry: an integer value representing the number of seconds since the epoch 00:00:00 UTC on 1 January 1970 at which the token will expire.
        :return: The signature portion of the Sas Token.

        :raises: IoTEdgeError if data cannot be signed
        """
        string_to_sign = quoted_resource_uri + "\n" + str(expiry)
        return self.hsm.sign(string_to_sign)


class IoTEdgeHsm(object):
    """
    Constructor for instantiating a iot hsm object.  This is an object that
    communicates with the Azure IoT Edge HSM in order to get connection credentials
    for an Azure IoT Edge module.  The credentials that this object return come in
    two forms:

    1. The trust bundle, which is a certificate that can be used as a trusted cert
       to authenticate the SSL connection between the IoE Edge module and IoT Edge
    2. A signing function, which can be used to create the sig field for a
       SharedAccessSignature string which can be used to authenticate with Iot Edge
    """

    def __init__(self, module_id, module_generation_id, workload_uri, api_version):
        """
        Constructor for instantiating a Azure IoT Edge HSM object

        :param str module_id: The module id
        :param str api_version: The API version
        :param str module_generation_id: The module generation id
        :param str workload_uri: The workload uri
        """
        self.module_id = urllib.parse.quote(module_id)
        self.api_version = api_version
        self.module_generation_id = module_generation_id
        self.workload_uri = _format_socket_uri(workload_uri)

    # TODO: Is this really the right name? It returns a certificate FROM the trust bundle,
    # not the trust bundle itself
    def get_trust_bundle(self):
        """
        Return the trust bundle that can be used to validate the server-side SSL
        TLS connection that we use to talk to edgeHub.

        :return: The server verification certificate to use for connections to the Azure IoT Edge
        instance, as a PEM certificate in string form.

        :raises: IoTEdgeError if unable to retrieve the certificate.
        """
        r = requests.get(
            self.workload_uri + "trust-bundle",
            params={"api-version": self.api_version},
            headers={"User-Agent": urllib.parse.quote_plus(ProductInfo.get_iothub_user_agent())},
        )
        # Validate that the request was successful
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise IoTEdgeError(message="Unable to get trust bundle from EdgeHub", cause=e)
        # Decode the trust bundle
        try:
            bundle = r.json()
        except ValueError as e:
            raise IoTEdgeError(message="Unable to decode trust bundle", cause=e)
        # Retrieve the certificate
        try:
            cert = bundle["certificate"]
        except KeyError as e:
            raise IoTEdgeError(message="No certificate in trust bundle", cause=e)
        return cert

    def sign(self, data_str):
        """
        Use the IoTEdge HSM to sign a piece of string data.  The caller should then insert the
        returned value (the signature) into the 'sig' field of a SharedAccessSignature string.

        :param str data_str: The data string to sign

        :return: The signature, as a URI-encoded and base64-encoded value that is ready to
        directly insert into the SharedAccessSignature string.

        :raises: IoTEdgeError if unable to sign the data.
        """
        encoded_data_str = base64.b64encode(data_str.encode("utf-8")).decode()

        path = (
            self.workload_uri
            + "modules/"
            + self.module_id
            + "/genid/"
            + self.module_generation_id
            + "/sign"
        )
        sign_request = {"keyId": "primary", "algo": "HMACSHA256", "data": encoded_data_str}

        r = requests.post(  # TODO: can we use json field instead of data?
            url=path,
            params={"api-version": self.api_version},
            headers={"User-Agent": urllib.parse.quote_plus(ProductInfo.get_iothub_user_agent())},
            data=json.dumps(sign_request),
        )
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise IoTEdgeError(message="Unable to sign data", cause=e)
        try:
            sign_response = r.json()
        except ValueError as e:
            raise IoTEdgeError(message="Unable to decode signed data", cause=e)
        try:
            signed_data_str = sign_response["digest"]
        except KeyError as e:
            raise IoTEdgeError(message="No signed data received", cause=e)

        return urllib.parse.quote(signed_data_str)


def _format_socket_uri(old_uri):
    """
    This function takes a socket URI in one form and converts it into another form.

    The source form is based on what we receive inside the IOTEDGE_WORKLOADURI
    environment variable, and it looks like this:
    "unix:///var/run/iotedge/workload.sock"

    The destination form is based on what the requests_unixsocket library expects
    and it looks like this:
    "http+unix://%2Fvar%2Frun%2Fiotedge%2Fworkload.sock/"

    The function changes the prefix, uri-encodes the path, and adds a slash
    at the end.

    If the socket URI does not start with unix:// this function only adds
    a slash at the end.

    :param old_uri: The URI in IOTEDGE_WORKLOADURI form

    :return: The URI in requests_unixsocket form
    """
    old_prefix = "unix://"
    new_prefix = "http+unix://"

    if old_uri.startswith(old_prefix):
        stripped_uri = old_uri[len(old_prefix) :]
        if stripped_uri.endswith("/"):
            stripped_uri = stripped_uri[:-1]
        new_uri = new_prefix + urllib.parse.quote(stripped_uri, safe="")
    else:
        new_uri = old_uri

    if not new_uri.endswith("/"):
        new_uri += "/"

    return new_uri



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import json
import logging
from azure.iot.device.common.pipeline import (
    pipeline_events_base,
    pipeline_ops_base,
    PipelineStage,
    pipeline_thread,
)
from azure.iot.device import exceptions
from azure.iot.device.common import handle_exceptions
from azure.iot.device.common.callable_weak_method import CallableWeakMethod
from . import pipeline_events_iothub, pipeline_ops_iothub
from . import constant

logger = logging.getLogger(__name__)


class UseAuthProviderStage(PipelineStage):
    """
    PipelineStage which extracts relevant AuthenticationProvider values for a new
    SetIoTHubConnectionArgsOperation.

    All other operations are passed down.
    """

    def __init__(self):
        super(UseAuthProviderStage, self).__init__()
        self.auth_provider = None

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_iothub.SetAuthProviderOperation):
            self.auth_provider = op.auth_provider
            # Here we append rather than just add it to the handler value because otherwise it
            # would overwrite the handler from another pipeline that might be using the same auth provider.
            self.auth_provider.on_sas_token_updated_handler_list.append(
                CallableWeakMethod(self, "_on_sas_token_updated")
            )
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_iothub.SetIoTHubConnectionArgsOperation,
                device_id=self.auth_provider.device_id,
                module_id=self.auth_provider.module_id,
                hostname=self.auth_provider.hostname,
                gateway_hostname=getattr(self.auth_provider, "gateway_hostname", None),
                server_verification_cert=getattr(
                    self.auth_provider, "server_verification_cert", None
                ),
                sas_token=self.auth_provider.get_current_sas_token(),
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_iothub.SetX509AuthProviderOperation):
            self.auth_provider = op.auth_provider
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_iothub.SetIoTHubConnectionArgsOperation,
                device_id=self.auth_provider.device_id,
                module_id=self.auth_provider.module_id,
                hostname=self.auth_provider.hostname,
                gateway_hostname=getattr(self.auth_provider, "gateway_hostname", None),
                server_verification_cert=getattr(
                    self.auth_provider, "server_verification_cert", None
                ),
                client_cert=self.auth_provider.get_x509_certificate(),
            )
            self.send_op_down(worker_op)
        else:
            super(UseAuthProviderStage, self)._run_op(op)

    @pipeline_thread.invoke_on_pipeline_thread_nowait
    def _on_sas_token_updated(self):
        logger.info(
            "{}: New sas token received.  Passing down UpdateSasTokenOperation.".format(self.name)
        )

        @pipeline_thread.runs_on_pipeline_thread
        def on_token_update_complete(op, error):
            if error:
                logger.error(
                    "{}({}): token update operation failed.  Error={}".format(
                        self.name, op.name, error
                    )
                )
                handle_exceptions.handle_background_exception(error)
            else:
                logger.debug(
                    "{}({}): token update operation is complete".format(self.name, op.name)
                )

        self.send_op_down(
            pipeline_ops_base.UpdateSasTokenOperation(
                sas_token=self.auth_provider.get_current_sas_token(),
                callback=on_token_update_complete,
            )
        )


class EnsureDesiredPropertiesStage(PipelineStage):
    """
    Pipeline stage Responsible for making sure that desired properties are always kept up to date.
    It does this by sending diwn a GetTwinOperation after a connection is reestablished, and, if
    the desired properties have changed since the last time a patch was received, it will send up
    an artificial patch event to send those updated properties to the app.
    """

    def __init__(self):
        self.last_version_seen = None
        self.pending_get_request = None
        super(EnsureDesiredPropertiesStage, self).__init__()

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_base.EnableFeatureOperation):
            # If we're enabling twin patches, we set last_version_seen to -1
            # as a way of enabling this functionality.  If the ConnectedEvent handler
            # sees this -1, it will send a GetTwinOperation to refresh desired properties.

            if op.feature_name == constant.TWIN_PATCHES:
                logger.info(
                    "{}: enabling twin patches.  setting last_version_seen".format(self.name)
                )
                self.last_version_seen = -1
        self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _ensure_get_op(self):
        """
        Function which makes sure we have a GetTwin operation in progress.  If we've
        already sent one down and we're waiting for it to return, we don't want to send
        a new one down.  This is because layers below us (especially CoordinateRequestAndResponseStage)
        will do everything they can to ensure we get a response on the already-pending
        GetTwinOperation.
        """
        if not self.pending_get_request:
            logger.info("{}: sending twin GET to ensure freshness".format(self.name))
            self.pending_get_request = pipeline_ops_iothub.GetTwinOperation(
                callback=CallableWeakMethod(self, "_on_get_twin_complete")
            )
            self.send_op_down(self.pending_get_request)
        else:
            logger.info(
                "{}: Outstanding twin GET already exists.  Not sending anything".format(self.name)
            )

    @pipeline_thread.runs_on_pipeline_thread
    def _on_get_twin_complete(self, op, error):
        """
        Function that gets called when a GetTwinOperation _that_we_initiated_ is complete.
        This is where we compare $version values and decide if we want to create an artificial
        TwinDesiredPropertiesPatchEvent or not.
        """

        logger.info("{}: _on_twin_get_complete".format(self.name))
        self.pending_get_request = None
        if error:
            # If the GetTwinOperation failed, we blindly try again.  We run the risk of
            # repeating this forever and might need to add logic to "give up" after some
            # number of failures, but we don't have any real reason to add that just yet.

            logger.info("{}: Twin GET failed with error {}.  Resubmitting.".format(self, error))
            self._ensure_get_op()
        else:
            logger.info("{} Twin GET response received.  Checking versions".format(self))
            new_version = op.twin["desired"]["$version"]
            logger.info(
                "{}: old version = {}, new version = {}".format(
                    self.name, self.last_version_seen, new_version
                )
            )
            if self.last_version_seen != new_version:
                # The twin we received has different (presumably newer) desired properties.
                # Make an artificial patch and send it up

                logger.info("{}: Version changed.  Sending up new patch event".format(self.name))
                self.last_version_seen = new_version
                self.send_event_up(
                    pipeline_events_iothub.TwinDesiredPropertiesPatchEvent(op.twin["desired"])
                )

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        if isinstance(event, pipeline_events_iothub.TwinDesiredPropertiesPatchEvent):
            # remember the $version when we get a patch.
            version = event.patch["$version"]
            logger.info(
                "{}: Desired patch received.  Saving $version={}".format(self.name, version)
            )
            self.last_version_seen = version
        elif isinstance(event, pipeline_events_base.ConnectedEvent):
            # If last_version_seen is truthy, that means we've seen desired property patches
            # before (or we've enabled them at least).  If this is the case, get the twin to
            # see if the desired props have been updated.
            if self.last_version_seen:
                logger.info("{}: Reconnected.  Getting twin")
                self._ensure_get_op()
        self.send_event_up(event)


class TwinRequestResponseStage(PipelineStage):
    """
    PipelineStage which handles twin operations. In particular, it converts twin GET and PATCH
    operations into RequestAndResponseOperation operations.  This is done at the IoTHub level because
    there is nothing protocol-specific about this code.  The protocol-specific implementation
    for twin requests and responses is handled inside IoTHubMQTTTranslationStage, when it converts
    the RequestOperation to a protocol-specific send operation and when it converts the
    protocol-specific receive event into an ResponseEvent event.
    """

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        def map_twin_error(error, twin_op):
            if error:
                return error
            elif twin_op.status_code >= 300:
                # TODO map error codes to correct exceptions
                logger.error("Error {} received from twin operation".format(twin_op.status_code))
                logger.error("response body: {}".format(twin_op.response_body))
                return exceptions.ServiceError(
                    "twin operation returned status {}".format(twin_op.status_code)
                )

        if isinstance(op, pipeline_ops_iothub.GetTwinOperation):

            # Alias to avoid overload within the callback below
            # CT-TODO: remove the need for this with better callback semantics
            op_waiting_for_response = op

            def on_twin_response(op, error):
                logger.debug("{}({}): Got response for GetTwinOperation".format(self.name, op.name))
                error = map_twin_error(error=error, twin_op=op)
                if not error:
                    op_waiting_for_response.twin = json.loads(op.response_body.decode("utf-8"))
                op_waiting_for_response.complete(error=error)

            self.send_op_down(
                pipeline_ops_base.RequestAndResponseOperation(
                    request_type=constant.TWIN,
                    method="GET",
                    resource_location="/",
                    request_body=" ",
                    callback=on_twin_response,
                )
            )

        elif isinstance(op, pipeline_ops_iothub.PatchTwinReportedPropertiesOperation):

            # Alias to avoid overload within the callback below
            # CT-TODO: remove the need for this with better callback semantics
            op_waiting_for_response = op

            def on_twin_response(op, error):
                logger.debug(
                    "{}({}): Got response for PatchTwinReportedPropertiesOperation operation".format(
                        self.name, op.name
                    )
                )
                error = map_twin_error(error=error, twin_op=op)
                op_waiting_for_response.complete(error=error)

            logger.debug(
                "{}({}): Sending reported properties patch: {}".format(self.name, op.name, op.patch)
            )

            self.send_op_down(
                pipeline_ops_base.RequestAndResponseOperation(
                    request_type=constant.TWIN,
                    method="PATCH",
                    resource_location="/properties/reported/",
                    request_body=json.dumps(op.patch),
                    callback=on_twin_response,
                )
            )

        else:
            super(TwinRequestResponseStage, self)._run_op(op)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.iot.device.common.pipeline import PipelineOperation


class MethodInvokeOperation(PipelineOperation):
    """
    A PipleineOperation object which contains arguments used to send a method invoke to an IoTHub or EdgeHub server.

    This operation is in the group of EdgeHub operations because it is very specific to the EdgeHub client.
    """

    def __init__(self, target_device_id, target_module_id, method_params, callback):
        """
        Initializer for MethodInvokeOperation objects.

        :param str target_device_id: The device id of the target device/module
        :param str target_module_id: The module id of the target module
        :param method_params: The parameters used to invoke the method, as defined by the IoT Hub specification.
        :param callback: The function that gets called when this operation is complete or has failed.
            The callback function must accept a PipelineOperation object which indicates the specific operation has which
            has completed or failed.
        :type callback: Function/callable
        """
        super(MethodInvokeOperation, self).__init__(callback=callback)
        self.target_device_id = target_device_id
        self.target_module_id = target_module_id
        self.method_params = method_params
        self.method_response = None


class GetStorageInfoOperation(PipelineOperation):
    """
    A PipleineOperation object which contains arguments used to get the storage information from IoT Hub.
    """

    def __init__(self, blob_name, callback):
        """
        Initializer for GetStorageInfo objects.

        :param str blob_name: The name of the blob that will be created in Azure Storage
        :param callback: The function that gets called when this operation is complete or has failed.
            The callback function must accept a PipelineOperation object which indicates the specific operation has which
            has completed or failed.
        :type callback: Function/callable

        :ivar dict storage_info: Upon completion, this contains the storage information which was retrieved from the service.
        """
        super(GetStorageInfoOperation, self).__init__(callback=callback)
        self.blob_name = blob_name
        self.storage_info = None


class NotifyBlobUploadStatusOperation(PipelineOperation):
    """
    A PipleineOperation object which contains arguments used to get the storage information from IoT Hub.
    """

    def __init__(self, correlation_id, is_success, status_code, status_description, callback):
        """
        Initializer for GetStorageInfo objects.

        :param str correlation_id: Provided by IoT Hub on get_storage_info_for_blob request.
        :param bool is_success: A boolean that indicates whether the file was uploaded successfully.
        :param int request_status_code: A numeric status code that is the status for the upload of the fiel to storage.
        :param str status_description: A description that corresponds to the status_code.
        :param callback: The function that gets called when this operation is complete or has failed.
            The callback function must accept a PipelineOperation object which indicates the specific operation has which
            has completed or failed.
        :type callback: Function/callable
        """
        super(NotifyBlobUploadStatusOperation, self).__init__(callback=callback)
        self.correlation_id = correlation_id
        self.is_success = is_success
        self.request_status_code = status_code
        self.status_description = status_description



"""Azure Provisioning Device Security

This package provides security and authentication-related functionality for use with the
Azure Provisioning Device SDK.
"""

from .sk_security_client import SymmetricKeySecurityClient
from .x509_security_client import X509SecurityClient



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from . import PipelineOperation


class SetMQTTConnectionArgsOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to connect to a server using the MQTT protocol.

    This operation is in the group of MQTT operations because its attributes are very specific to the MQTT protocol.
    """

    def __init__(
        self,
        client_id,
        hostname,
        username,
        callback,
        server_verification_cert=None,
        client_cert=None,
        sas_token=None,
    ):
        """
        Initializer for SetMQTTConnectionArgsOperation objects.

        :param str client_id: The client identifier to use when connecting to the MQTT server
        :param str hostname: The hostname of the MQTT server we will eventually connect to
        :param str username: The username to use when connecting to the MQTT server
        :param str server_verification_cert: (Optional) The server verification certificate to use
          if the MQTT server that we're going to connect to uses server-side TLS
        :param X509 client_cert: (Optional) The x509 object containing a client certificate and key used to connect
          to the MQTT service
        :param str sas_token: The token string which will be used to authenticate with the service
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super(SetMQTTConnectionArgsOperation, self).__init__(callback=callback)
        self.client_id = client_id
        self.hostname = hostname
        self.username = username
        self.server_verification_cert = server_verification_cert
        self.client_cert = client_cert
        self.sas_token = sas_token


class MQTTPublishOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to publish a specific payload on a specific topic using the MQTT protocol.

    This operation is in the group of MQTT operations because its attributes are very specific to the MQTT protocol.
    """

    def __init__(self, topic, payload, callback):
        """
        Initializer for MQTTPublishOperation objects.

        :param str topic: The name of the topic to publish to
        :param str payload: The payload to publish
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super(MQTTPublishOperation, self).__init__(callback=callback)
        self.topic = topic
        self.payload = payload
        self.needs_connection = True
        self.retry_timer = None


class MQTTSubscribeOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to subscribe to a specific MQTT topic using the MQTT protocol.

    This operation is in the group of MQTT operations because its attributes are very specific to the MQTT protocol.
    """

    def __init__(self, topic, callback):
        """
        Initializer for MQTTSubscribeOperation objects.

        :param str topic: The name of the topic to subscribe to
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super(MQTTSubscribeOperation, self).__init__(callback=callback)
        self.topic = topic
        self.needs_connection = True
        self.timeout_timer = None
        self.retry_timer = None


class MQTTUnsubscribeOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to unsubscribe from a specific MQTT topic using the MQTT protocol.

    This operation is in the group of MQTT operations because its attributes are very specific to the MQTT protocol.
    """

    def __init__(self, topic, callback):
        """
        Initializer for MQTTUnsubscribeOperation objects.

        :param str topic: The name of the topic to unsubscribe from
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super(MQTTUnsubscribeOperation, self).__init__(callback=callback)
        self.topic = topic
        self.needs_connection = True
        self.timeout_timer = None
        self.retry_timer = None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains abstract classes for the various clients of the Azure IoT Hub Device SDK
"""

import six
import abc
import logging
import os
import io
from . import auth
from . import pipeline

logger = logging.getLogger(__name__)

# A note on implementation:
# The intializer methods accept pipeline(s) instead of an auth provider in order to protect
# the client from logic related to authentication providers. This reduces edge cases, and allows
# pipeline configuration to be specifically tailored to the method of instantiation.
# For instance, .create_from_connection_string and .create_from_edge_envrionment both can use
# SymmetricKeyAuthenticationProviders to instantiate pipeline(s), but only .create_from_edge_environment
# should use it to instantiate an HTTPPipeline. If the initializer accepted an auth provider, and then
# used it to create pipelines, this detail would be lost, as there would be no way to tell if a
# SymmetricKeyAuthenticationProvider was intended to be part of an Edge scenario or not.


def _validate_kwargs(**kwargs):
    """Helper function to validate user provided kwargs.
    Raises TypeError if an invalid option has been provided"""
    valid_kwargs = [
        "product_info",
        "websockets",
        "cipher",
        "server_verification_cert",
        "proxy_options",
    ]

    for kwarg in kwargs:
        if kwarg not in valid_kwargs:
            raise TypeError("Got an unexpected keyword argument '{}'".format(kwarg))


def _get_pipeline_config_kwargs(**kwargs):
    """Helper function to get a subset of user provided kwargs relevant to IoTHubPipelineConfig"""
    new_kwargs = {}
    if "product_info" in kwargs:
        new_kwargs["product_info"] = kwargs["product_info"]
    if "websockets" in kwargs:
        new_kwargs["websockets"] = kwargs["websockets"]
    if "cipher" in kwargs:
        new_kwargs["cipher"] = kwargs["cipher"]
    if "proxy_options" in kwargs:
        new_kwargs["proxy_options"] = kwargs["proxy_options"]
    return new_kwargs


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubClient(object):
    """ A superclass representing a generic IoTHub client.
    This class needs to be extended for specific clients.
    """

    def __init__(self, mqtt_pipeline, http_pipeline):
        """Initializer for a generic client.

        :param mqtt_pipeline: The pipeline used to connect to the IoTHub endpoint.
        :type mqtt_pipeline: :class:`azure.iot.device.iothub.pipeline.MQTTPipeline`
        """
        self._mqtt_pipeline = mqtt_pipeline
        self._http_pipeline = http_pipeline

    @classmethod
    def create_from_connection_string(cls, connection_string, **kwargs):
        """
        Instantiate the client from a IoTHub device or module connection string.

        :param str connection_string: The connection string for the IoTHub you wish to connect to.

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`

        :raises: ValueError if given an invalid connection_string.
        :raises: TypeError if given an unrecognized parameter.

        :returns: An instance of an IoTHub client that uses a connection string for authentication.
        """
        # TODO: Make this device/module specific and reject non-matching connection strings.
        # This will require refactoring of the auth package to use common objects (e.g. ConnectionString)
        # in order to differentiate types of connection strings.

        _validate_kwargs(**kwargs)

        # Pipeline Config setup
        pipeline_config_kwargs = _get_pipeline_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(**pipeline_config_kwargs)
        if cls.__name__ == "IoTHubDeviceClient":
            pipeline_configuration.blob_upload = True

        # Auth Provider setup
        authentication_provider = auth.SymmetricKeyAuthenticationProvider.parse(connection_string)
        authentication_provider.server_verification_cert = kwargs.get("server_verification_cert")

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(authentication_provider, pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(authentication_provider, pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @abc.abstractmethod
    def connect(self):
        pass

    @abc.abstractmethod
    def disconnect(self):
        pass

    @abc.abstractmethod
    def send_message(self, message):
        pass

    @abc.abstractmethod
    def receive_method_request(self, method_name=None):
        pass

    @abc.abstractmethod
    def send_method_response(self, method_request, payload, status):
        pass

    @abc.abstractmethod
    def get_twin(self):
        pass

    @abc.abstractmethod
    def patch_twin_reported_properties(self, reported_properties_patch):
        pass

    @abc.abstractmethod
    def receive_twin_desired_properties_patch(self):
        pass

    @property
    def connected(self):
        """
        Read-only property to indicate if the transport is connected or not.
        """
        return self._mqtt_pipeline.connected


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubDeviceClient(AbstractIoTHubClient):
    @classmethod
    def create_from_x509_certificate(cls, x509, hostname, device_id, **kwargs):
        """
        Instantiate a client which using X509 certificate authentication.

        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param x509: The complete x509 certificate object.
            To use the certificate the enrollment object needs to contain cert
            (either the root certificate or one of the intermediate CA certificates).
            If the cert comes from a CER file, it needs to be base64 encoded.
        :type x509: :class:`azure.iot.device.X509`
        :param str device_id: The ID used to uniquely identify a device in the IoTHub

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`

        :raises: TypeError if given an unrecognized parameter.

        :returns: An instance of an IoTHub client that uses an X509 certificate for authentication.
        """
        _validate_kwargs(**kwargs)

        # Pipeline Config setup
        pipeline_config_kwargs = _get_pipeline_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(**pipeline_config_kwargs)
        pipeline_configuration.blob_upload = True  # Blob Upload is a feature on Device Clients

        # Auth Provider setup
        authentication_provider = auth.X509AuthenticationProvider(
            x509=x509, hostname=hostname, device_id=device_id
        )
        authentication_provider.server_verification_cert = kwargs.get("server_verification_cert")

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(authentication_provider, pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(authentication_provider, pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @classmethod
    def create_from_symmetric_key(cls, symmetric_key, hostname, device_id, **kwargs):
        """
        Instantiate a client using symmetric key authentication.

        :param symmetric_key: The symmetric key.
        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param device_id: The device ID

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`

        :raises: TypeError if given an unrecognized parameter.

        :return: An instance of an IoTHub client that uses a symmetric key for authentication.
        """
        _validate_kwargs(**kwargs)

        # Pipeline Config setup
        pipeline_config_kwargs = _get_pipeline_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(**pipeline_config_kwargs)
        pipeline_configuration.blob_upload = True  # Blob Upload is a feature on Device Clients

        # Auth Provider setup
        authentication_provider = auth.SymmetricKeyAuthenticationProvider(
            hostname=hostname, device_id=device_id, module_id=None, shared_access_key=symmetric_key
        )
        authentication_provider.server_verification_cert = kwargs.get("server_verification_cert")

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(authentication_provider, pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(authentication_provider, pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @abc.abstractmethod
    def receive_message(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubModuleClient(AbstractIoTHubClient):
    def __init__(self, mqtt_pipeline, http_pipeline):
        """Initializer for a module client.

        :param mqtt_pipeline: The pipeline used to connect to the IoTHub endpoint.
        :type mqtt_pipeline: :class:`azure.iot.device.iothub.pipeline.MQTTPipeline`
        """
        super(AbstractIoTHubModuleClient, self).__init__(mqtt_pipeline, http_pipeline)

    @classmethod
    def create_from_edge_environment(cls, **kwargs):
        """
        Instantiate the client from the IoT Edge environment.

        This method can only be run from inside an IoT Edge container, or in a debugging
        environment configured for Edge development (e.g. Visual Studio, Visual Studio Code)

        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`

        :raises: OSError if the IoT Edge container is not configured correctly.
        :raises: ValueError if debug variables are invalid.

        :returns: An instance of an IoTHub client that uses the IoT Edge environment for
            authentication.
        """
        _validate_kwargs(**kwargs)
        if kwargs.get("server_verification_cert"):
            raise TypeError(
                "'server_verification_cert' is not supported by clients using an IoT Edge environment"
            )

        # First try the regular Edge container variables
        try:
            hostname = os.environ["IOTEDGE_IOTHUBHOSTNAME"]
            device_id = os.environ["IOTEDGE_DEVICEID"]
            module_id = os.environ["IOTEDGE_MODULEID"]
            gateway_hostname = os.environ["IOTEDGE_GATEWAYHOSTNAME"]
            module_generation_id = os.environ["IOTEDGE_MODULEGENERATIONID"]
            workload_uri = os.environ["IOTEDGE_WORKLOADURI"]
            api_version = os.environ["IOTEDGE_APIVERSION"]
        except KeyError:
            # As a fallback, try the Edge local dev variables for debugging.
            # These variables are set by VS/VS Code in order to allow debugging
            # of Edge application code in a non-Edge dev environment.
            try:
                connection_string = os.environ["EdgeHubConnectionString"]
                ca_cert_filepath = os.environ["EdgeModuleCACertificateFile"]
            except KeyError as e:
                new_err = OSError("IoT Edge environment not configured correctly")
                new_err.__cause__ = e
                raise new_err
            # TODO: variant server_verification_cert file vs data object that would remove the need for this fopen
            # Read the certificate file to pass it on as a string
            try:
                with io.open(ca_cert_filepath, mode="r") as ca_cert_file:
                    server_verification_cert = ca_cert_file.read()
            except (OSError, IOError) as e:
                # In Python 2, a non-existent file raises IOError, and an invalid file raises an IOError.
                # In Python 3, a non-existent file raises FileNotFoundError, and an invalid file raises an OSError.
                # However, FileNotFoundError inherits from OSError, and IOError has been turned into an alias for OSError,
                # thus we can catch the errors for both versions in this block.
                # Unfortunately, we can't distinguish cause of error from error type, so the raised ValueError has a generic
                # message. If, in the future, we want to add detail, this could be accomplished by inspecting the e.errno
                # attribute
                new_err = ValueError("Invalid CA certificate file")
                new_err.__cause__ = e
                raise new_err
            # Use Symmetric Key authentication for local dev experience.
            try:
                authentication_provider = auth.SymmetricKeyAuthenticationProvider.parse(
                    connection_string
                )
            except ValueError:
                raise
            authentication_provider.server_verification_cert = server_verification_cert
        else:
            # Use an HSM for authentication in the general case
            try:
                authentication_provider = auth.IoTEdgeAuthenticationProvider(
                    hostname=hostname,
                    device_id=device_id,
                    module_id=module_id,
                    gateway_hostname=gateway_hostname,
                    module_generation_id=module_generation_id,
                    workload_uri=workload_uri,
                    api_version=api_version,
                )
            except auth.IoTEdgeError as e:
                new_err = OSError("Unexpected failure in IoTEdge")
                new_err.__cause__ = e
                raise new_err

        # Pipeline Config setup
        pipeline_config_kwargs = _get_pipeline_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(**pipeline_config_kwargs)
        pipeline_configuration.method_invoke = (
            True
        )  # Method Invoke is allowed on modules created from edge environment

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(authentication_provider, pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(authentication_provider, pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @classmethod
    def create_from_x509_certificate(cls, x509, hostname, device_id, module_id, **kwargs):
        """
        Instantiate a client which using X509 certificate authentication.

        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param x509: The complete x509 certificate object.
            To use the certificate the enrollment object needs to contain cert
            (either the root certificate or one of the intermediate CA certificates).
            If the cert comes from a CER file, it needs to be base64 encoded.
        :type x509: :class:`azure.iot.device.X509`
        :param str device_id: The ID used to uniquely identify a device in the IoTHub
        :param str module_id: The ID used to uniquely identify a module on a device on the IoTHub.

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`

        :raises: TypeError if given an unrecognized parameter.

        :returns: An instance of an IoTHub client that uses an X509 certificate for authentication.
        """
        _validate_kwargs(**kwargs)

        # Pipeline Config setup
        pipeline_config_kwargs = _get_pipeline_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(**pipeline_config_kwargs)

        # Auth Provider setup
        authentication_provider = auth.X509AuthenticationProvider(
            x509=x509, hostname=hostname, device_id=device_id, module_id=module_id
        )
        authentication_provider.server_verification_cert = kwargs.get("server_verification_cert")

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(authentication_provider, pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(authentication_provider, pipeline_configuration)
        return cls(mqtt_pipeline, http_pipeline)

    @abc.abstractmethod
    def send_message_to_output(self, message, output_name):
        pass

    @abc.abstractmethod
    def receive_message_on_input(self, input_name):
        pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains classes related to device twin and module twin functionality
"""


class Twin(object):
    """Represents a device twin or module twin

    :ivar desired_properties: The desired properties for the Twin.  These are properties
      which are sent _to_ the device or module to indicate the _desired_ state of the device
      or module
    :type desired_properties: dict, str, int, float, bool, or None (JSON compatible values)
    :ivar reported_properties: The reported properties for the Twin.  These are properties
      which are sent _from_ the device or module to indicate the _actual_ state of the device.
    :type reported_properties: dict, str, int, float, bool, or None (JSON compatible values)
    """

    # OPEN QUESTION: can desired_properties and reported_properties only be dict, or can they be any JSON-compatible value?

    def __init__(self):
        """Initializer for a Twin object
        """
        self.desiried_properties = None
        self.reported_properties = None



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains a client that is responsible for providing x509 certificates that will eventually establish
 the authenticity of devices to Device Provisioning Service.
"""


class X509SecurityClient(object):
    """
    An X509 Security Client. This uses the certificate and key provided to authenticate a device
    with an Azure DPS instance.X509 Authentication is only supported for device identities
    connecting directly to an Azure DPS.
    """

    def __init__(self, provisioning_host, registration_id, id_scope, x509):
        """
        Initialize the X509 Certificate security client.
        :param provisioning_host: Host running the Device Provisioning Service. Can be found in the Azure portal in the
        Overview tab as the string Global device endpoint
        :param registration_id: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        The registration ID is alphanumeric, lowercase string and may contain hyphens.
        :param id_scope: The ID scope is used to uniquely identify the specific provisioning service the device will
        register through. The ID scope is assigned to a Device Provisioning Service when it is created by the user and
        is generated by the service and is immutable, guaranteeing uniqueness.
        :param x509: The x509 certificate, To use the certificate the enrollment object needs to contain cert (either the root certificate or one of the intermediate CA certificates).
        If the cert comes from a CER file, it needs to be base64 encoded.
        """
        self._provisioning_host = provisioning_host
        self._registration_id = registration_id
        self._id_scope = id_scope
        self._x509 = x509

    @property
    def provisioning_host(self):
        """
        :return: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        The registration ID is alphanumeric, lowercase string and may contain hyphens.
        """
        return self._provisioning_host

    @property
    def registration_id(self):
        """
        :return: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        The registration ID is alphanumeric, lowercase string and may contain hyphens.
        """
        return self._registration_id

    @property
    def id_scope(self):
        """
        :return: Host running the Device Provisioning Service.
        """
        return self._id_scope

    def get_x509_certificate(self):
        """
        :return: The x509 certificate, To use the certificate the enrollment object needs to contain
         cert (either the root certificate or one of the intermediate CA certificates).
        """
        return self._x509



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from azure.iot.device.common.pipeline.config import BasePipelineConfig

logger = logging.getLogger(__name__)


class ProvisioningPipelineConfig(BasePipelineConfig):
    """A class for storing all configurations/options for Provisioning clients in the Azure IoT Python Device Client Library.
    """

    pass



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.iot.device.common.pipeline import PipelineOperation


# TODO: Combine SetAuthProviderOperation and SetX509AuthProviderOperation once
# auth provider is reduced to a simple vector
class SetX509AuthProviderOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to use a particular X509 authorization provider.
    Some pipeline stage is expected to extract arguments out of the auth provider and pass them
    on so an even lower stage can use those arguments to connect.

    This operation is in the group of IoTHub operations because authorization providers are currently
    very IoTHub-specific
    """

    def __init__(self, auth_provider, callback):
        """
        Initializer for SetAuthProviderOperation objects.

        :param object auth_provider: The X509 authorization provider object to use to retrieve connection parameters
          which can be used to connect to the service.
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super(SetX509AuthProviderOperation, self).__init__(callback=callback)
        self.auth_provider = auth_provider


class SetAuthProviderOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to use a particular authorization provider.
    Some pipeline stage is expected to extract arguments out of the auth provider and pass them
    on so an even lower stage can use those arguments to connect.

    This operation is in the group of IoTHub operations because autorization providers are currently
    very IoTHub-specific
    """

    def __init__(self, auth_provider, callback):
        """
        Initializer for SetAuthProviderOperation objects.

        :param object auth_provider: The authorization provider object to use to retrieve connection parameters
          which can be used to connect to the service.
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super(SetAuthProviderOperation, self).__init__(callback=callback)
        self.auth_provider = auth_provider


class SetIoTHubConnectionArgsOperation(PipelineOperation):
    """
    A PipelineOperation object which contains connection arguments which were retrieved from an authorization provider,
    likely by a pipeline stage which handles the SetAuthProviderOperation operation.

    This operation is in the group of IoTHub operations because the arguments which it accepts are very specific to
    IoTHub connections and would not apply to other types of client connections (such as a DPS client).
    """

    def __init__(
        self,
        device_id,
        hostname,
        callback,
        module_id=None,
        gateway_hostname=None,
        server_verification_cert=None,
        client_cert=None,
        sas_token=None,
    ):
        """
        Initializer for SetIoTHubConnectionArgsOperation objects.

        :param str device_id: The device id for the device that we are connecting.
        :param str hostname: The hostname of the iothub service we are connecting to.
        :param str module_id: (optional) If we are connecting as a module, this contains the module id
          for the module we are connecting.
        :param str gateway_hostname: (optional) If we are going through a gateway host, this is the
          hostname for the gateway
        :param str server_verification_cert: (Optional) The server verification certificate to use
          if the server that we're going to connect to uses server-side TLS
        :param X509 client_cert: (Optional) The x509 object containing a client certificate and key used to connect
          to the service
        :param str sas_token: The token string which will be used to authenticate with the service
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super(SetIoTHubConnectionArgsOperation, self).__init__(callback=callback)
        self.device_id = device_id
        self.module_id = module_id
        self.hostname = hostname
        self.gateway_hostname = gateway_hostname
        self.server_verification_cert = server_verification_cert
        self.client_cert = client_cert
        self.sas_token = sas_token


class SendD2CMessageOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to send a telemetry message to an IoTHub or EdegHub server.

    This operation is in the group of IoTHub operations because it is very specific to the IoTHub client
    """

    def __init__(self, message, callback):
        """
        Initializer for SendD2CMessageOperation objects.

        :param Message message: The message that we're sending to the service
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super(SendD2CMessageOperation, self).__init__(callback=callback)
        self.message = message


class SendOutputEventOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to send an output message to an EdgeHub server.

    This operation is in the group of IoTHub operations because it is very specific to the IoTHub client
    """

    def __init__(self, message, callback):
        """
        Initializer for SendOutputEventOperation objects.

        :param Message message: The output message that we're sending to the service. The name of the output is
          expected to be stored in the output_name attribute of this object
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super(SendOutputEventOperation, self).__init__(callback=callback)
        self.message = message


class SendMethodResponseOperation(PipelineOperation):
    """
    A PipleineOperation object which contains arguments used to send a method response to an IoTHub or EdgeHub server.

    This operation is in the group of IoTHub operations because it is very specific to the IoTHub client.
    """

    def __init__(self, method_response, callback):
        """
        Initializer for SendMethodResponseOperation objects.

        :param method_response: The method response to be sent to IoTHub/EdgeHub
        :type method_response: MethodResponse
        :param callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept a PipelineOperation object which indicates the specific operation has which
         has completed or failed.
        :type callback: Function/callable
        """
        super(SendMethodResponseOperation, self).__init__(callback=callback)
        self.method_response = method_response


class GetTwinOperation(PipelineOperation):
    """
    A PipelineOperation object which represents a request to get a device twin or a module twin from an Azure
    IoT Hub or Azure Iot Edge Hub service.

    :ivar twin: Upon completion, this contains the twin which was retrieved from the service.
    :type twin: Twin
    """

    def __init__(self, callback):
        """
        Initializer for GetTwinOperation objects.
        """
        super(GetTwinOperation, self).__init__(callback=callback)
        self.twin = None


class PatchTwinReportedPropertiesOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to send a reported properties patch to the Azure
    IoT Hub or Azure IoT Edge Hub service.
    """

    def __init__(self, patch, callback):
        """
        Initializer for PatchTwinReportedPropertiesOperation object

        :param patch: The reported properties patch to send to the service.
        :type patch: dict, str, int, float, bool, or None (JSON compatible values)
        """
        super(PatchTwinReportedPropertiesOperation, self).__init__(callback=callback)
        self.patch = patch



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import six
import abc
from azure.iot.device.common import models

logger = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class BasePipelineConfig(object):
    """A base class for storing all configurations/options shared across the Azure IoT Python Device Client Library.
    More specific configurations such as those that only apply to the IoT Hub Client will be found in the respective
    config files.
    """

    def __init__(self, websockets=False, cipher="", proxy_options=None):
        """Initializer for BasePipelineConfig

        :param bool websockets: Enabling/disabling websockets in MQTT. This feature is relevant
            if a firewall blocks port 8883 from use.
        :param cipher: Optional cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param proxy_options: Details of proxy configuration
        :type proxy_options: :class:`azure.iot.device.common.models.ProxyOptions`
        """
        self.websockets = websockets
        self.cipher = self._sanitize_cipher(cipher)
        self.proxy_options = proxy_options

    @staticmethod
    def _sanitize_cipher(cipher):
        """Sanitize the cipher input and convert to a string in OpenSSL list format
        """
        if isinstance(cipher, list):
            cipher = ":".join(cipher)

        if isinstance(cipher, str):
            cipher = cipher.upper()
            cipher = cipher.replace("_", "-")
        else:
            raise TypeError("Invalid type for 'cipher'")

        return cipher



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains a client that is responsible for providing shared access tokens that will eventually establish
 the authenticity of devices to Device Provisioning Service.
"""

from azure.iot.device.common.sastoken import SasToken


class SymmetricKeySecurityClient(object):
    """
    A client that is responsible for providing shared access tokens that will eventually establish
    the authenticity of devices to Device Provisioning Service.
    :ivar provisioning_host: Host running the Device Provisioning Service
    :ivar registration_id: : The registration ID is used to uniquely identify a device in the Device Provisioning Service.
    :ivar id_scope: : The ID scope is used to uniquely identify the specific provisioning service the device will
        register through.
    """

    def __init__(self, provisioning_host, registration_id, id_scope, symmetric_key):
        """
        Initialize the symmetric key security client.
        :param provisioning_host: Host running the Device Provisioning Service. Can be found in the Azure portal in the
        Overview tab as the string Global device endpoint
        :param registration_id: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        The registration ID is alphanumeric, lowercase string and may contain hyphens.
        :param id_scope: The ID scope is used to uniquely identify the specific provisioning service the device will
        register through. The ID scope is assigned to a Device Provisioning Service when it is created by the user and
        is generated by the service and is immutable, guaranteeing uniqueness.
        :param symmetric_key: The key which will be used to create the shared access signature token to authenticate
        the device with the Device Provisioning Service. By default, the Device Provisioning Service creates
        new symmetric keys with a default length of 32 bytes when new enrollments are saved with the Auto-generate keys
        option enabled. Users can provide their own symmetric keys for enrollments by disabling this option within
        16 bytes and 64 bytes and in valid Base64 format.
        """
        self._provisioning_host = provisioning_host
        self._registration_id = registration_id
        self._id_scope = id_scope
        self._symmetric_key = symmetric_key
        self._sas_token = None

    @property
    def provisioning_host(self):
        """
        :return: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        The registration ID is alphanumeric, lowercase string and may contain hyphens.
        """
        return self._provisioning_host

    @property
    def registration_id(self):
        """
        :return: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        The registration ID is alphanumeric, lowercase string and may contain hyphens.
        """
        return self._registration_id

    @property
    def id_scope(self):
        """
        :return: Host running the Device Provisioning Service.
        """
        return self._id_scope

    def _create_shared_access_signature(self):
        """
        Construct SAS tokens that have a hashed signature formed using the symmetric key of this security client.
        This signature is recreated by the Device Provisioning Service to verify whether a security token presented
        during attestation is authentic or not.
        :return: A string representation of the shared access signature which is of the form
        SharedAccessSignature sig={signature}&se={expiry}&skn={policyName}&sr={URL-encoded-resourceURI}
        """
        uri = self._id_scope + "/registrations/" + self._registration_id
        key = self._symmetric_key
        time_to_live = 3600
        keyname = "registration"
        return SasToken(uri, key, keyname, time_to_live)

    def get_current_sas_token(self):
        if self._sas_token is None:
            self._sas_token = self._create_shared_access_signature()
        else:
            self._sas_token.refresh()
        return str(self._sas_token)



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


class ChainableException(Exception):
    """This exception stores a reference to a previous exception which has caused
    the current one"""

    def __init__(self, message=None, cause=None):
        # By using .__cause__, this will allow typical stack trace behavior in Python 3,
        # while still being able to operate in Python 2.
        self.__cause__ = cause
        super(ChainableException, self).__init__(message)

    def __str__(self):
        if self.__cause__:
            return "{} caused by {}".format(
                super(ChainableException, self).__repr__(), self.__cause__.__repr__()
            )
        else:
            return super(ChainableException, self).__repr__()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import six.moves.urllib as urllib
from azure.iot.device.common import version_compat
from azure.iot.device.common.pipeline import (
    pipeline_ops_base,
    pipeline_ops_mqtt,
    pipeline_events_mqtt,
    pipeline_thread,
    pipeline_events_base,
)
from azure.iot.device.common.pipeline.pipeline_stages_base import PipelineStage
from azure.iot.device.provisioning.pipeline import mqtt_topic_provisioning
from azure.iot.device.provisioning.pipeline import pipeline_ops_provisioning
from azure.iot.device import constant as pkg_constant
from . import constant as pipeline_constant
from azure.iot.device.product_info import ProductInfo

logger = logging.getLogger(__name__)


class ProvisioningMQTTTranslationStage(PipelineStage):
    """
    PipelineStage which converts other Provisioning pipeline operations into MQTT operations. This stage also
    converts MQTT pipeline events into Provisioning pipeline events.
    """

    def __init__(self):
        super(ProvisioningMQTTTranslationStage, self).__init__()
        self.action_to_topic = {}

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):

        if isinstance(op, pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation):
            # get security client args from above, save some, use some to build topic names,
            # always pass it down because MQTT protocol stage will also want to receive these args.

            client_id = op.registration_id
            query_param_seq = [
                ("api-version", pkg_constant.PROVISIONING_API_VERSION),
                ("ClientVersion", ProductInfo.get_provisioning_user_agent()),
            ]
            username = "{id_scope}/registrations/{registration_id}/{query_params}".format(
                id_scope=op.id_scope,
                registration_id=op.registration_id,
                query_params=version_compat.urlencode(
                    query_param_seq, quote_via=urllib.parse.quote
                ),
            )

            hostname = op.provisioning_host

            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.SetMQTTConnectionArgsOperation,
                client_id=client_id,
                hostname=hostname,
                username=username,
                client_cert=op.client_cert,
                sas_token=op.sas_token,
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_base.RequestOperation):
            if op.request_type == pipeline_constant.REGISTER:
                topic = mqtt_topic_provisioning.get_register_topic_for_publish(
                    request_id=op.request_id
                )
                worker_op = op.spawn_worker_op(
                    worker_op_type=pipeline_ops_mqtt.MQTTPublishOperation,
                    topic=topic,
                    payload=op.request_body,
                )
                self.send_op_down(worker_op)
            else:
                topic = mqtt_topic_provisioning.get_query_topic_for_publish(
                    request_id=op.request_id, operation_id=op.query_params["operation_id"]
                )
                worker_op = op.spawn_worker_op(
                    worker_op_type=pipeline_ops_mqtt.MQTTPublishOperation,
                    topic=topic,
                    payload=op.request_body,
                )
                self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_base.EnableFeatureOperation):
            # Enabling for register gets translated into an MQTT subscribe operation
            topic = mqtt_topic_provisioning.get_register_topic_for_subscribe()
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.MQTTSubscribeOperation, topic=topic
            )
            self.send_op_down(worker_op)

        elif isinstance(op, pipeline_ops_base.DisableFeatureOperation):
            # Disabling a register response gets turned into an MQTT unsubscribe operation
            topic = mqtt_topic_provisioning.get_register_topic_for_subscribe()
            worker_op = op.spawn_worker_op(
                worker_op_type=pipeline_ops_mqtt.MQTTUnsubscribeOperation, topic=topic
            )
            self.send_op_down(worker_op)

        else:
            # All other operations get passed down
            super(ProvisioningMQTTTranslationStage, self)._run_op(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        """
        Pipeline Event handler function to convert incoming MQTT messages into the appropriate DPS
        events, based on the topic of the message
        """
        if isinstance(event, pipeline_events_mqtt.IncomingMQTTMessageEvent):
            topic = event.topic

            if mqtt_topic_provisioning.is_dps_response_topic(topic):
                logger.info(
                    "Received payload:{payload} on topic:{topic}".format(
                        payload=event.payload, topic=topic
                    )
                )
                key_values = mqtt_topic_provisioning.extract_properties_from_dps_response_topic(
                    topic
                )
                retry_after = key_values.get("retry-after", None)
                status_code = mqtt_topic_provisioning.extract_status_code_from_dps_response_topic(
                    topic
                )
                request_id = key_values["rid"]

                self.send_event_up(
                    pipeline_events_base.ResponseEvent(
                        request_id=request_id,
                        status_code=int(status_code, 10),
                        response_body=event.payload,
                        retry_after=retry_after,
                    )
                )
            else:
                logger.warning("Unknown topic: {} passing up to next handler".format(topic))
                self.send_event_up(event)

        else:
            # all other messages get passed up
            super(ProvisioningMQTTTranslationStage, self)._handle_pipeline_event(event)



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import json


class RegistrationResult(object):
    """
    The final result of a completed or failed registration attempt
    :ivar:request_id: The request id to which the response is being obtained
    :ivar:operation_id: The id of the operation as returned by the registration request.
    :ivar status: The status of the registration process as returned by the provisioning service.
    Values can be "unassigned", "assigning", "assigned", "failed", "disabled"
    :ivar registration_state : Details like device id, assigned hub , date times etc returned
    from the provisioning service.
    """

    def __init__(self, operation_id, status, registration_state=None):
        """
        :param operation_id: The id of the operation as returned by the initial registration request.
        :param status: The status of the registration process.
        Values can be "unassigned", "assigning", "assigned", "failed", "disabled"
        :param registration_state : Details like device id, assigned hub , date times etc returned
        from the provisioning service.
        """
        self._operation_id = operation_id
        self._status = status
        self._registration_state = registration_state

    @property
    def operation_id(self):
        return self._operation_id

    @property
    def status(self):
        return self._status

    @property
    def registration_state(self):
        return self._registration_state

    def __str__(self):
        return "\n".join([str(self.registration_state), self.status])


class RegistrationState(object):
    """
    The registration state regarding the device.
    :ivar device_id: Desired device id for the provisioned device
    :ivar assigned_hub: Desired IoT Hub to which the device is linked.
    :ivar sub_status: Substatus for 'Assigned' devices. Possible values are
    "initialAssignment", "deviceDataMigrated", "deviceDataReset"
    :ivar created_date_time: Registration create date time (in UTC).
    :ivar last_update_date_time: Last updated date time (in UTC).
    :ivar etag: The entity tag associated with the resource.
    """

    def __init__(
        self,
        device_id=None,
        assigned_hub=None,
        sub_status=None,
        created_date_time=None,
        last_update_date_time=None,
        etag=None,
        payload=None,
    ):
        """
        :param device_id: Desired device id for the provisioned device
        :param assigned_hub: Desired  IoT Hub to which the device is linked.
        :param sub_status: Substatus for 'Assigned' devices. Possible values are
        "initialAssignment", "deviceDataMigrated", "deviceDataReset"
        :param created_date_time: Registration create date time (in UTC).
        :param last_update_date_time: Last updated date time (in UTC).
        :param etag: The entity tag associated with the resource.
        :param payload: The payload with which hub is responding
        """
        self._device_id = device_id
        self._assigned_hub = assigned_hub
        self._sub_status = sub_status
        self._created_date_time = created_date_time
        self._last_update_date_time = last_update_date_time
        self._etag = etag
        self._response_payload = payload

    @property
    def device_id(self):
        return self._device_id

    @property
    def assigned_hub(self):
        return self._assigned_hub

    @property
    def sub_status(self):
        return self._sub_status

    @property
    def created_date_time(self):
        return self._created_date_time

    @property
    def last_update_date_time(self):
        return self._last_update_date_time

    @property
    def etag(self):
        return self._etag

    @property
    def response_payload(self):
        return json.dumps(self._response_payload, default=lambda o: o.__dict__, sort_keys=True)

    def __str__(self):
        return "\n".join(
            [self.device_id, self.assigned_hub, self.sub_status, self.response_payload]
        )
